"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import base64
import logging
import struct
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    overload,
)

from discord_protos import PreloadedUserSettings  # , FrecencyUserSettings
from google.protobuf.json_format import MessageToDict, ParseDict

from .activity import CustomActivity
from .colour import Colour
from .enums import (
    EmojiPickerSection,
    HighlightLevel,
    InboxTab,
    Locale,
    NotificationLevel,
    SpoilerRenderOptions,
    Status,
    StickerAnimationOptions,
    StickerPickerSection,
    Theme,
    UserContentFilter,
    try_enum,
)
from .flags import (
    FriendDiscoveryFlags,
    FriendSourceFlags,
    HubProgressFlags,
    OnboardingProgressFlags,
)
from .object import Object
from .utils import (
    MISSING,
    _get_as_snowflake,
    _ocast,
    find,
    parse_time,
    parse_timestamp,
    utcnow,
)

if TYPE_CHECKING:
    from google.protobuf.message import Message
    from typing_extensions import Self

    from .abc import GuildChannel, Snowflake
    from .channel import DMChannel, GroupChannel
    from .guild import Guild
    from .state import ConnectionState
    from .user import ClientUser, User

    PrivateChannel = Union[DMChannel, GroupChannel]

__all__ = (
    "UserSettings",
    "GuildFolder",
    "GuildProgress",
    "AudioContext",
    "LegacyUserSettings",
    "MuteConfig",
    "ChannelSettings",
    "GuildSettings",
    "TrackingSettings",
    "EmailSettings",
)

_log = logging.getLogger(__name__)


class _ProtoSettings:
    __slots__ = (
        "_state",
        "settings",
    )

    PROTOBUF_CLS: Type[Message] = MISSING
    settings: Any

    # I honestly wish I didn't have to vomit properties everywhere like this,
    # but unfortunately it's probably the best way to do it
    # The discord-protos library is maintained seperately, so any changes
    # to the protobufs will have to be reflected here;
    # this is why I'm keeping the `settings` attribute public
    # I love protobufs :blobcatcozystars:

    def __init__(self, state: ConnectionState, data: str):
        self._state: ConnectionState = state
        self._update(data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.settings == other.settings
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.settings != other.settings
        return True

    def _update(self, data: str, *, partial: bool = False):
        if partial:
            self.merge_from_base64(data)
        else:
            self.from_base64(data)

    @classmethod
    def _copy(cls, self: Self, /) -> Self:
        new = cls.__new__(cls)
        new._state = self._state
        new.settings = cls.PROTOBUF_CLS()
        new.settings.CopyFrom(self.settings)
        return new

    @overload
    def _get_guild(self, id: int, /, *, always_guild: Literal[True] = ...) -> Guild: ...

    @overload
    def _get_guild(
        self, id: int, /, *, always_guild: Literal[False] = ...
    ) -> Union[Guild, Object]: ...

    def _get_guild(
        self, id: int, /, *, always_guild: bool = False
    ) -> Union[Guild, Object]:
        id = int(id)
        if always_guild:
            return self._state._get_or_create_unavailable_guild(id)
        return self._state._get_guild(id) or Object(id=id)

    def to_dict(self, *, with_defaults: bool = False) -> Dict[str, Any]:
        return MessageToDict(
            self.settings,
            including_default_value_fields=with_defaults,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )

    def dict_to_base64(self, data: Dict[str, Any]) -> str:
        message = ParseDict(data, self.PROTOBUF_CLS())
        return base64.b64encode(message.SerializeToString()).decode("ascii")

    def from_base64(self, data: str):
        self.settings = self.PROTOBUF_CLS().FromString(base64.b64decode(data))

    def merge_from_base64(self, data: str):
        self.settings.MergeFromString(base64.b64decode(data))

    def to_base64(self) -> str:
        return base64.b64encode(self.settings.SerializeToString()).decode("ascii")


class UserSettings(_ProtoSettings):
    """Represents the Discord client settings.

    .. versionadded:: 2.0
    """

    __slots__ = ()

    PROTOBUF_CLS = PreloadedUserSettings

    # Client versions are supposed to be backwards compatible
    # If the client supports a version newer than the one in data,
    # it does a migration and updates the version in data
    SUPPORTED_CLIENT_VERSION = 17
    SUPPORTED_SERVER_VERSION = 0

    def __init__(self, *args):
        super().__init__(*args)
        if self.client_version < self.SUPPORTED_CLIENT_VERSION:
            # Migrations are mostly for client state, but we'll throw a debug log anyway
            _log.debug(
                "PreloadedUserSettings client version is outdated, migration needed. Unexpected behaviour may occur."
            )
        if self.server_version > self.SUPPORTED_SERVER_VERSION:
            # At the time of writing, the server version is not provided (so it's always 0)
            # The client does not use the field at all, so there probably won't be any server-side migrations anytime soon
            _log.debug(
                "PreloadedUserSettings server version is newer than supported. Unexpected behaviour may occur."
            )

    @property
    def data_version(self) -> int:
        """:class:`int`: The version of the settings. Increases on every change."""
        return self.settings.versions.data_version

    @property
    def client_version(self) -> int:
        """:class:`int`: The client version of the settings. Used for client-side data migrations."""
        return self.settings.versions.client_version

    @property
    def server_version(self) -> int:
        """:class:`int`: The server version of the settings. Used for server-side data migrations."""
        return self.settings.versions.server_version

    # Inbox Settings

    @property
    def inbox_tab(self) -> InboxTab:
        """:class:`InboxTab`: The current (last opened) inbox tab."""
        return try_enum(InboxTab, self.settings.inbox.current_tab)

    @property
    def inbox_tutorial_viewed(self) -> bool:
        """:class:`bool`: Whether the inbox tutorial has been viewed."""
        return self.settings.inbox.viewed_tutorial

    # Guild Settings

    @property
    def guild_progress_settings(self) -> List[GuildProgress]:
        """List[:class:`GuildProgress`]: A list of guild progress settings."""
        state = self._state
        return [
            GuildProgress._from_settings(guild_id, data=settings, state=state)
            for guild_id, settings in self.settings.guilds.guilds.items()
        ]

    # User Content Settings

    @property
    def dismissed_contents(self) -> Tuple[int, ...]:
        """Tuple[:class:`int`]: A list of enum values representing dismissable content in the app.

        .. note::

            For now, this just returns the raw values without converting to a proper enum,
            as the enum values change too often to be viably maintained.
        """
        contents = self.settings.user_content.dismissed_contents
        return struct.unpack(f">{len(contents)}B", contents)

    @property
    def last_dismissed_promotion_start_date(self) -> Optional[datetime]:
        """Optional[:class:`datetime.datetime`]: The date the last dismissed promotion started."""
        return parse_time(
            self.settings.user_content.last_dismissed_outbound_promotion_start_date.value
            or None
        )

    @property
    def nitro_basic_modal_dismissed_at(self) -> Optional[datetime]:
        """Optional[:class:`datetime.datetime`]: The date the Nitro Basic modal was dismissed."""
        return (
            self.settings.user_content.premium_tier_0_modal_dismissed_at.ToDatetime(
                tzinfo=timezone.utc
            )
            if self.settings.user_content.HasField("premium_tier_0_modal_dismissed_at")
            else None
        )

    # Voice and Video Settings

    # TODO: Video filters

    # @property
    # def video_filter_background_blur(self) -> bool:
    #     return self.settings.voice_and_video.blur.use_blur

    @property
    def always_preview_video(self) -> bool:
        """Whether to always show the preview modal when the user turns on their camera."""
        return self.settings.voice_and_video.always_preview_video.value

    @property
    def afk_timeout(self) -> int:
        """:class:`int`: How long (in seconds) the user needs to be AFK until Discord sends push notifications to mobile devices (30-600)."""
        return self.settings.voice_and_video.afk_timeout.value or 600

    @property
    def stream_notifications_enabled(self) -> bool:
        """:class:`bool`: Whether stream notifications for friends will be received."""
        return (
            self.settings.voice_and_video.stream_notifications_enabled.value
            if self.settings.voice_and_video.HasField("stream_notifications_enabled")
            else True
        )

    @property
    def native_phone_integration_enabled(self) -> bool:
        """:class:`bool`: Whether to enable the Discord mobile Callkit."""
        return (
            self.settings.voice_and_video.native_phone_integration_enabled.value
            if self.settings.voice_and_video.HasField(
                "native_phone_integration_enabled"
            )
            else True
        )

    @property
    def soundboard_volume(self) -> float:
        """:class:`float`: The volume of the soundboard (0-100)."""
        return (
            self.settings.voice_and_video.soundboard_settings.volume
            if self.settings.voice_and_video.HasField("soundboard_settings")
            else 100.0
        )

    # Text and Images Settings

    @property
    def diversity_surrogate(self) -> Optional[str]:
        """Optional[:class:`str`]: The unicode character used as the diversity surrogate for supported emojis (i.e. emoji skin tones, ``🏻``)."""
        return self.settings.text_and_images.diversity_surrogate.value or None

    @property
    def use_thread_sidebar(self) -> bool:
        """:class:`bool`: Whether to open threads in split view."""
        return (
            self.settings.text_and_images.use_thread_sidebar.value
            if self.settings.text_and_images.HasField("use_thread_sidebar")
            else True
        )

    @property
    def render_spoilers(self) -> SpoilerRenderOptions:
        """:class:`SpoilerRenderOptions`: When to show spoiler content."""
        return try_enum(
            SpoilerRenderOptions,
            self.settings.text_and_images.render_spoilers.value or "ON_CLICK",
        )

    @property
    def collapsed_emoji_picker_sections(
        self,
    ) -> Tuple[Union[EmojiPickerSection, Guild], ...]:
        """Tuple[Union[:class:`EmojiPickerSection`, :class:`Guild`]]: A list of emoji picker sections (including guild IDs) that are collapsed."""
        return tuple(
            self._get_guild(section, always_guild=True)
            if section.isdigit()
            else try_enum(EmojiPickerSection, section)
            for section in self.settings.text_and_images.emoji_picker_collapsed_sections
        )

    @property
    def collapsed_sticker_picker_sections(
        self,
    ) -> Tuple[Union[StickerPickerSection, Guild, Object], ...]:
        """Tuple[Union[:class:`StickerPickerSection`, :class:`Guild`, :class:`Object`]]: A list of sticker picker sections (including guild and sticker pack IDs) that are collapsed."""
        return tuple(
            self._get_guild(section, always_guild=False)
            if section.isdigit()
            else try_enum(StickerPickerSection, section)
            for section in self.settings.text_and_images.sticker_picker_collapsed_sections
        )

    @property
    def view_image_descriptions(self) -> bool:
        """:class:`bool`: Whether to display the alt text of attachments."""
        return self.settings.text_and_images.view_image_descriptions.value

    @property
    def show_command_suggestions(self) -> bool:
        """:class:`bool`: Whether to show application command suggestions in-chat."""
        return (
            self.settings.text_and_images.show_command_suggestions.value
            if self.settings.text_and_images.HasField("show_command_suggestions")
            else True
        )

    @property
    def inline_attachment_media(self) -> bool:
        """:class:`bool`: Whether to display attachments when they are uploaded in chat."""
        return (
            self.settings.text_and_images.inline_attachment_media.value
            if self.settings.text_and_images.HasField("inline_attachment_media")
            else True
        )

    @property
    def inline_embed_media(self) -> bool:
        """:class:`bool`: Whether to display videos and images from links posted in chat."""
        return (
            self.settings.text_and_images.inline_embed_media.value
            if self.settings.text_and_images.HasField("inline_embed_media")
            else True
        )

    @property
    def gif_auto_play(self) -> bool:
        """:class:`bool`: Whether to automatically play GIFs that are in the chat.."""
        return (
            self.settings.text_and_images.gif_auto_play.value
            if self.settings.text_and_images.HasField("gif_auto_play")
            else True
        )

    @property
    def render_embeds(self) -> bool:
        """:class:`bool`: Whether to render embeds that are sent in the chat."""
        return (
            self.settings.text_and_images.render_embeds.value
            if self.settings.text_and_images.HasField("render_embeds")
            else True
        )

    @property
    def render_reactions(self) -> bool:
        """:class:`bool`: Whether to render reactions that are added to messages."""
        return (
            self.settings.text_and_images.render_reactions.value
            if self.settings.text_and_images.HasField("render_reactions")
            else True
        )

    @property
    def animate_emojis(self) -> bool:
        """:class:`bool`: Whether to animate emojis in the chat."""
        return (
            self.settings.text_and_images.animate_emoji.value
            if self.settings.text_and_images.HasField("animate_emoji")
            else True
        )

    @property
    def animate_stickers(self) -> StickerAnimationOptions:
        """:class:`StickerAnimationOptions`: Whether to animate stickers in the chat."""
        return try_enum(
            StickerAnimationOptions,
            self.settings.text_and_images.animate_stickers.value,
        )

    @property
    def enable_tts_command(self) -> bool:
        """:class:`bool`: Whether to allow TTS messages to be played/sent."""
        return (
            self.settings.text_and_images.enable_tts_command.value
            if self.settings.text_and_images.HasField("enable_tts_command")
            else True
        )

    @property
    def message_display_compact(self) -> bool:
        """:class:`bool`: Whether to use the compact Discord display mode."""
        return self.settings.text_and_images.message_display_compact.value

    @property
    def explicit_content_filter(self) -> UserContentFilter:
        """:class:`UserContentFilter`: The filter for explicit content in all messages."""
        return try_enum(
            UserContentFilter,
            self.settings.text_and_images.explicit_content_filter.value
            if self.settings.text_and_images.HasField("explicit_content_filter")
            else 1,
        )

    @property
    def view_nsfw_guilds(self) -> bool:
        """:class:`bool`: Whether to show NSFW guilds on iOS."""
        return self.settings.text_and_images.view_nsfw_guilds.value

    @property
    def convert_emoticons(self) -> bool:
        r""":class:`bool`: Whether to automatically convert emoticons into emojis (e.g. ``:)`` -> 😃)."""
        return (
            self.settings.text_and_images.convert_emoticons.value
            if self.settings.text_and_images.HasField("convert_emoticons")
            else True
        )

    @property
    def show_expression_suggestions(self) -> bool:
        """:class:`bool`: Whether to show expression (emoji/sticker/soundboard) suggestions in-chat."""
        return (
            self.settings.text_and_images.expression_suggestions_enabled.value
            if self.settings.text_and_images.HasField("expression_suggestions_enabled")
            else True
        )

    @property
    def view_nsfw_commands(self) -> bool:
        """:class:`bool`: Whether to show NSFW application commands in DMs."""
        return self.settings.text_and_images.view_nsfw_commands.value

    @property
    def use_legacy_chat_input(self) -> bool:
        """:class:`bool`: Whether to use the legacy chat input over the new rich input."""
        return self.settings.text_and_images.use_legacy_chat_input.value

    # Notifications Settings

    @property
    def in_app_notifications(self) -> bool:
        """:class:`bool`: Whether to show notifications directly in the app."""
        return (
            self.settings.notifications.show_in_app_notifications.value
            if self.settings.notifications.HasField("show_in_app_notifications")
            else True
        )

    @property
    def send_stream_notifications(self) -> bool:
        """:class:`bool`: Whether to send notifications to friends when using the go live feature."""
        return self.settings.notifications.notify_friends_on_go_live.value

    @property
    def notification_center_acked_before_id(self) -> int:
        """:class:`int`: The ID of the last notification that was acknowledged in the notification center."""
        return self.settings.notifications.notification_center_acked_before_id

    # Privacy Settings

    @property
    def allow_activity_friend_joins(self) -> bool:
        """:class:`bool`: Whether to allow friends to join your activity without sending a request."""
        return (
            self.settings.privacy.allow_activity_party_privacy_friends.value
            if self.settings.privacy.HasField("allow_activity_party_privacy_friends")
            else True
        )

    @property
    def allow_activity_voice_channel_joins(self) -> bool:
        """:class:`bool`: Whether to allow people in the same voice channel as you to join your activity without sending a request. Does not apply to Community guilds."""
        return (
            self.settings.privacy.allow_activity_party_privacy_voice_channel.value
            if self.settings.privacy.HasField(
                "allow_activity_party_privacy_voice_channel"
            )
            else True
        )

    @property
    def restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that you will not receive DMs from."""
        return list(map(self._get_guild, self.settings.privacy.restricted_guild_ids))

    @property
    def default_guilds_restricted(self) -> bool:
        """:class:`bool`: Whether to automatically disable DMs between you and members of new guilds you join."""
        return self.settings.privacy.default_guilds_restricted

    @property
    def allow_accessibility_detection(self) -> bool:
        """:class:`bool`: Whether to allow Discord to track screen reader usage."""
        return self.settings.privacy.allow_accessibility_detection

    @property
    def detect_platform_accounts(self) -> bool:
        """:class:`bool`: Whether to automatically detect accounts from services like Steam and Blizzard when you open the Discord client."""
        return (
            self.settings.privacy.detect_platform_accounts.value
            if self.settings.privacy.HasField("detect_platform_accounts")
            else True
        )

    @property
    def passwordless(self) -> bool:
        """:class:`bool`: Whether to enable passwordless login."""
        return (
            self.settings.privacy.passwordless.value
            if self.settings.privacy.HasField("passwordless")
            else True
        )

    @property
    def contact_sync_enabled(self) -> bool:
        """:class:`bool`: Whether to enable the contact sync on Discord mobile."""
        return self.settings.privacy.contact_sync_enabled.value

    @property
    def friend_source_flags(self) -> FriendSourceFlags:
        """:class:`FriendSourceFlags`: Who can add you as a friend."""
        return (
            FriendSourceFlags._from_value(
                self.settings.privacy.friend_source_flags.value
            )
            if self.settings.privacy.HasField("friend_source_flags")
            else FriendSourceFlags.all()
        )

    @property
    def friend_discovery_flags(self) -> FriendDiscoveryFlags:
        """:class:`FriendDiscoveryFlags`: How you get recommended friends."""
        return FriendDiscoveryFlags._from_value(
            self.settings.privacy.friend_discovery_flags.value
        )

    @property
    def activity_restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that your current activity will not be shown in."""
        return list(
            map(self._get_guild, self.settings.privacy.activity_restricted_guild_ids)
        )

    @property
    def default_guilds_activity_restricted(self) -> bool:
        """:class:`bool`: Whether to automatically disable showing your current activity in new large (over 200 member) guilds you join."""
        return self.settings.privacy.default_guilds_activity_restricted

    @property
    def activity_joining_restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that will not be able to join your current activity."""
        return list(
            map(
                self._get_guild,
                self.settings.privacy.activity_joining_restricted_guild_ids,
            )
        )

    @property
    def message_request_restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds whose originating DMs will not be filtered into your message requests."""
        return list(
            map(
                self._get_guild,
                self.settings.privacy.message_request_restricted_guild_ids,
            )
        )

    @property
    def default_message_request_restricted(self) -> bool:
        """:class:`bool`: Whether to automatically disable the message request system in new guilds you join."""
        return self.settings.privacy.default_message_request_restricted.value

    @property
    def drops(self) -> bool:
        """:class:`bool`: Whether the Discord drops feature is enabled."""
        return not self.settings.privacy.drops_opted_out.value

    @property
    def non_spam_retraining(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to help improve Discord spam models when marking messages as non-spam; staff only."""
        return (
            self.settings.privacy.non_spam_retraining_opt_in.value
            if self.settings.privacy.HasField("non_spam_retraining_opt_in")
            else None
        )

    # Debug Settings

    @property
    def rtc_panel_show_voice_states(self) -> bool:
        """:class:`bool`: Whether to show voice states in the RTC panel."""
        return self.settings.debug.rtc_panel_show_voice_states.value

    # Game Library Settings

    @property
    def install_shortcut_desktop(self) -> bool:
        """:class:`bool`: Whether to install a desktop shortcut for games."""
        return self.settings.game_library.install_shortcut_desktop.value

    @property
    def install_shortcut_start_menu(self) -> bool:
        """:class:`bool`: Whether to install a start menu shortcut for games."""
        return (
            self.settings.game_library.install_shortcut_start_menu.value
            if self.settings.game_library.HasField("install_shortcut_start_menu")
            else True
        )

    @property
    def disable_games_tab(self) -> bool:
        """:class:`bool`: Whether to disable the showing of the Games tab."""
        return self.settings.game_library.disable_games_tab.value

    # Status Settings

    @property
    def status(self) -> Status:
        """:class:`Status`: The configured status."""
        return try_enum(Status, self.settings.status.status.value or "unknown")

    @property
    def custom_activity(self) -> Optional[CustomActivity]:
        """:class:`CustomActivity`: The set custom activity."""
        return (
            CustomActivity._from_settings(
                data=self.settings.status.custom_status, state=self._state
            )
            if self.settings.status.HasField("custom_status")
            else None
        )

    @property
    def show_current_game(self) -> bool:
        """:class:`bool`: Whether to show the current game."""
        return (
            self.settings.status.show_current_game.value
            if self.settings.status.HasField("show_current_game")
            else True
        )

    # Localization Settings

    @property
    def locale(self) -> Locale:
        """:class:`Locale`: The :rfc:`3066` language identifier of the locale to use for the language of the Discord client."""
        return try_enum(Locale, self.settings.localization.locale.value or "en-US")

    @property
    def timezone_offset(self) -> int:
        """:class:`int`: The timezone offset from UTC to use (in minutes)."""
        return self.settings.localization.timezone_offset.value

    # Appearance Settings

    @property
    def theme(self) -> Theme:
        """:class:`Theme`: The overall theme of the Discord UI."""
        return Theme.from_int(self.settings.appearance.theme)

    @property
    def client_theme(self) -> Optional[Tuple[int, int, float]]:
        """Optional[Tuple[:class:`int`, :class:`int`, :class:`float`]]: The client theme settings, in order of primary color, gradient preset, and gradient angle."""
        return (
            (
                self.settings.appearance.client_theme_settings.primary_color.value,
                self.settings.appearance.client_theme_settings.background_gradient_preset_id.value,
                self.settings.appearance.client_theme_settings.background_gradient_angle.value,
            )
            if self.settings.appearance.HasField("client_theme_settings")
            else None
        )

    @property
    def developer_mode(self) -> bool:
        """:class:`bool`: Whether to enable developer mode."""
        return self.settings.appearance.developer_mode

    @property
    def disable_mobile_redesign(self) -> bool:
        """:class:`bool`: Whether to opt-out of the mobile redesign."""
        return self.settings.appearance.mobile_redesign_disabled

    # Guild Folder Settings

    @property
    def guild_folders(self) -> List[GuildFolder]:
        """List[:class:`GuildFolder`]: A list of guild folders."""
        state = self._state
        return [
            GuildFolder._from_settings(data=folder, state=state)
            for folder in self.settings.guild_folders.folders
        ]

    @property
    def guild_positions(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds in order of the guild/guild icons that are on the left hand side of the UI."""
        return list(map(self._get_guild, self.settings.guild_folders.guild_positions))

    # Favorites Settings

    # TODO: Favorites

    # Audio Settings

    @property
    def user_audio_settings(self) -> List[AudioContext]:
        """List[:class:`AudioContext`]: A list of audio context settings for users."""
        state = self._state
        return [
            AudioContext._from_settings(user_id, data=data, state=state)
            for user_id, data in self.settings.audio_context_settings.user.items()
        ]

    @property
    def stream_audio_settings(self) -> List[AudioContext]:
        """List[:class:`AudioContext`]: A list of audio context settings for streams."""
        state = self._state
        return [
            AudioContext._from_settings(stream_id, data=data, state=state)
            for stream_id, data in self.settings.audio_context_settings.stream.items()
        ]

    # Communities Settings

    @property
    def home_auto_navigation(self) -> bool:
        """:class:`bool`: Whether to automatically redirect to guild home for guilds that have not been accessed in a while."""
        return not self.settings.communities.disable_home_auto_nav.value

    @overload
    async def edit(self) -> Self: ...

    @overload
    async def edit(
        self,
        *,
        require_version: Union[bool, int] = False,
        client_version: int = ...,
        inbox_tab: InboxTab = ...,
        inbox_tutorial_viewed: bool = ...,
        guild_progress_settings: Sequence[GuildProgress] = ...,
        dismissed_contents: Sequence[int] = ...,
        last_dismissed_promotion_start_date: datetime = ...,
        nitro_basic_modal_dismissed_at: datetime = ...,
        soundboard_volume: float = ...,
        afk_timeout: int = ...,
        always_preview_video: bool = ...,
        native_phone_integration_enabled: bool = ...,
        stream_notifications_enabled: bool = ...,
        diversity_surrogate: Optional[str] = ...,
        render_spoilers: SpoilerRenderOptions = ...,
        collapsed_emoji_picker_sections: Sequence[
            Union[EmojiPickerSection, Snowflake]
        ] = ...,
        collapsed_sticker_picker_sections: Sequence[
            Union[StickerPickerSection, Snowflake]
        ] = ...,
        animate_emojis: bool = ...,
        animate_stickers: StickerAnimationOptions = ...,
        explicit_content_filter: UserContentFilter = ...,
        show_expression_suggestions: bool = ...,
        use_thread_sidebar: bool = ...,
        view_image_descriptions: bool = ...,
        show_command_suggestions: bool = ...,
        inline_attachment_media: bool = ...,
        inline_embed_media: bool = ...,
        gif_auto_play: bool = ...,
        render_embeds: bool = ...,
        render_reactions: bool = ...,
        enable_tts_command: bool = ...,
        message_display_compact: bool = ...,
        view_nsfw_guilds: bool = ...,
        convert_emoticons: bool = ...,
        view_nsfw_commands: bool = ...,
        use_legacy_chat_input: bool = ...,
        in_app_notifications: bool = ...,
        send_stream_notifications: bool = ...,
        notification_center_acked_before_id: int = ...,
        allow_activity_friend_joins: bool = ...,
        allow_activity_voice_channel_joins: bool = ...,
        friend_source_flags: FriendSourceFlags = ...,
        friend_discovery_flags: FriendDiscoveryFlags = ...,
        drops: bool = ...,
        non_spam_retraining: Optional[bool] = ...,
        restricted_guilds: Sequence[Snowflake] = ...,
        default_guilds_restricted: bool = ...,
        allow_accessibility_detection: bool = ...,
        detect_platform_accounts: bool = ...,
        passwordless: bool = ...,
        contact_sync_enabled: bool = ...,
        activity_restricted_guilds: Sequence[Snowflake] = ...,
        default_guilds_activity_restricted: bool = ...,
        activity_joining_restricted_guilds: Sequence[Snowflake] = ...,
        message_request_restricted_guilds: Sequence[Snowflake] = ...,
        default_message_request_restricted: bool = ...,
        rtc_panel_show_voice_states: bool = ...,
        install_shortcut_desktop: bool = ...,
        install_shortcut_start_menu: bool = ...,
        disable_games_tab: bool = ...,
        status: Status = ...,
        custom_activity: Optional[CustomActivity] = ...,
        show_current_game: bool = ...,
        locale: Locale = ...,
        timezone_offset: int = ...,
        theme: Theme = ...,
        client_theme: Optional[Tuple[int, int, float]] = ...,
        disable_mobile_redesign: bool = ...,
        developer_mode: bool = ...,
        guild_folders: Sequence[GuildFolder] = ...,
        guild_positions: Sequence[Snowflake] = ...,
        user_audio_settings: Collection[AudioContext] = ...,
        stream_audio_settings: Collection[AudioContext] = ...,
        home_auto_navigation: bool = ...,
    ) -> Self: ...

    async def edit(
        self, *, require_version: Union[bool, int] = False, **kwargs: Any
    ) -> Self:
        r"""|coro|

        Edits the current user's settings.

        .. note::

            Settings subsections are not idempotently updated. This means if you change one setting in a subsection\* on an outdated
            instance of :class:`UserSettings` then the other settings in that subsection\* will be reset to the value of the instance.

            When operating on the cached user settings (i.e. :attr:`Client.settings`), this should not be an issue. However, if you
            are operating on a fetched instance, consider using the ``require_version`` parameter to ensure you don't overwrite
            newer settings.

            Any field may be explicitly set to ``MISSING`` to reset it to the default value.

            \* A subsection is a group of settings that are stored in the same top-level protobuf message.
            Examples include Privacy, Text and Images, Voice and Video, etc.

        .. note::

            This method is ratelimited heavily. Updates should be batched together and sent at intervals.

            Infrequent actions do not need a delay. Frequent actions should be delayed by 10 seconds and batched.
            Automated actions (such as migrations or frecency updates) should be delayed by 30 seconds and batched.
            Daily actions (things that change often and are not meaningful, such as emoji frencency) should be delayed by 1 day and batched.

        Parameters
        ----------
        require_version: Union[:class:`bool`, :class:`int`]
            Whether to require the current version of the settings to be the same as the provided version.
            If this is ``True`` then the current version is used.
        \*\*kwargs
            The settings to edit. Refer to the :class:`UserSettings` properties for the valid fields. Unknown fields are ignored.

        Raises
        ------
        HTTPException
            Editing the settings failed.
        TypeError
            At least one setting is required to edit.

        Returns
        -------
        :class:`UserSettings`
            The edited settings. Note that this is a new instance and not the same as the cached instance as mentioned above.
        """
        # As noted above, entire sections MUST be sent, or they will be reset to default values
        # Conversely, we want to omit fields that the user requests to be set to default (by explicitly passing MISSING)
        # For this, we then remove fields set to MISSING from the payload in the payload construction at the end

        if not kwargs:
            raise TypeError("edit() missing at least 1 required keyword-only argument")

        # Only client_version should ever really be sent
        versions = {}
        for field in ("data_version", "client_version", "server_version"):
            if field in kwargs:
                versions[field] = kwargs.pop(field)

        inbox = {}
        if "inbox_tab" in kwargs:
            inbox["current_tab"] = _ocast(kwargs.pop("inbox_tab"), int)
        if "inbox_tutorial_viewed" in kwargs:
            inbox["viewed_tutorial"] = kwargs.pop("inbox_tutorial_viewed")

        guilds = {}
        if (
            "guild_progress_settings" in kwargs
            and kwargs["guild_progress_settings"] is not MISSING
        ):
            guilds["guilds"] = (
                {
                    guild.guild_id: guild.to_dict()
                    for guild in kwargs.pop("guild_progress_settings")
                }
                if kwargs["guild_progress_settings"] is not MISSING
                else MISSING
            )

        user_content = {}
        if "dismissed_contents" in kwargs:
            contents = kwargs.pop("dismissed_contents")
            user_content["dismissed_contents"] = (
                struct.pack(f">{len(contents)}B", *contents)
                if contents is not MISSING
                else MISSING
            )
        if "last_dismissed_promotion_start_date" in kwargs:
            user_content["last_dismissed_outbound_promotion_start_date"] = (
                kwargs.pop("last_dismissed_promotion_start_date").isoformat()
                if kwargs["last_dismissed_promotion_start_date"] is not MISSING
                else MISSING
            )
        if "nitro_basic_modal_dismissed_at" in kwargs:
            user_content["premium_tier_0_modal_dismissed_at"] = (
                kwargs.pop("nitro_basic_modal_dismissed_at").isoformat()
                if kwargs["nitro_basic_modal_dismissed_at"] is not MISSING
                else MISSING
            )

        voice_and_video = {}
        if "soundboard_volume" in kwargs:
            voice_and_video["soundboard_settings"] = (
                {"volume": kwargs.pop("soundboard_volume")}
                if kwargs["soundboard_volume"] is not MISSING
                else {}
            )
        for field in (
            "afk_timeout",
            "always_preview_video",
            "native_phone_integration_enabled",
            "stream_notifications_enabled",
        ):
            if field in kwargs:
                voice_and_video[field] = kwargs.pop(field)

        text_and_images = {}
        if "diversity_surrogate" in kwargs:
            text_and_images["diversity_surrogate"] = (
                kwargs.pop("diversity_surrogate") or ""
                if kwargs["diversity_surrogate"] is not MISSING
                else MISSING
            )
        if "render_spoilers" in kwargs:
            text_and_images["render_spoilers"] = _ocast(
                kwargs.pop("render_spoilers"), str
            )
        if "collapsed_emoji_picker_sections" in kwargs:
            text_and_images["emoji_picker_collapsed_sections"] = (
                [
                    str(getattr(x, "id", x))
                    for x in kwargs.pop("collapsed_emoji_picker_sections")
                ]
                if kwargs["collapsed_emoji_picker_sections"] is not MISSING
                else MISSING
            )
        if "collapsed_sticker_picker_sections" in kwargs:
            text_and_images["sticker_picker_collapsed_sections"] = (
                [
                    str(getattr(x, "id", x))
                    for x in kwargs.pop("collapsed_sticker_picker_sections")
                ]
                if kwargs["collapsed_sticker_picker_sections"] is not MISSING
                else MISSING
            )
        if "animate_emojis" in kwargs:
            text_and_images["animate_emoji"] = kwargs.pop("animate_emojis")
        if "animate_stickers" in kwargs:
            text_and_images["animate_stickers"] = _ocast(
                kwargs.pop("animate_stickers"), int
            )
        if "explicit_content_filter" in kwargs:
            text_and_images["explicit_content_filter"] = _ocast(
                kwargs.pop("explicit_content_filter"), int
            )
        if "show_expression_suggestions" in kwargs:
            text_and_images["expression_suggestions_enabled"] = kwargs.pop(
                "show_expression_suggestions"
            )
        for field in (
            "use_thread_sidebar",
            "view_image_descriptions",
            "show_command_suggestions",
            "inline_attachment_media",
            "inline_embed_media",
            "gif_auto_play",
            "render_embeds",
            "render_reactions",
            "enable_tts_command",
            "message_display_compact",
            "view_nsfw_guilds",
            "convert_emoticons",
            "view_nsfw_commands",
            "use_legacy_chat_input",
            "use_rich_chat_input",
        ):
            if field in kwargs:
                text_and_images[field] = kwargs.pop(field)

        notifications = {}
        if "in_app_notifications" in kwargs:
            notifications["show_in_app_notifications"] = kwargs.pop(
                "in_app_notifications"
            )
        if "send_stream_notifications" in kwargs:
            notifications["notify_friends_on_go_live"] = kwargs.pop(
                "send_stream_notifications"
            )
        for field in ("notification_center_acked_before_id",):
            if field in kwargs:
                notifications[field] = kwargs.pop(field)

        privacy = {}
        if "allow_activity_friend_joins" in kwargs:
            privacy["allow_activity_party_privacy_friends"] = kwargs.pop(
                "allow_activity_friend_joins"
            )
        if "allow_activity_voice_channel_joins" in kwargs:
            privacy["allow_activity_party_privacy_voice_channel"] = kwargs.pop(
                "allow_activity_voice_channel_joins"
            )
        if "friend_source_flags" in kwargs:
            privacy["friend_source_flags"] = (
                kwargs.pop("friend_source_flags").value
                if kwargs["friend_source_flags"] is not MISSING
                else MISSING
            )
        if "friend_discovery_flags" in kwargs:
            privacy["friend_discovery_flags"] = (
                kwargs.pop("friend_discovery_flags").value
                if kwargs["friend_discovery_flags"] is not MISSING
                else MISSING
            )
        if "drops" in kwargs:
            privacy["drops_opted_out"] = (
                not kwargs.pop("drops") if kwargs["drops"] is not MISSING else MISSING
            )
        if "non_spam_retraining" in kwargs:
            privacy["non_spam_retraining_opt_in"] = (
                kwargs.pop("non_spam_retraining")
                if kwargs["non_spam_retraining"] not in {None, MISSING}
                else MISSING
            )
        for field in (
            "restricted_guilds",
            "default_guilds_restricted",
            "allow_accessibility_detection",
            "detect_platform_accounts",
            "passwordless",
            "contact_sync_enabled",
            "activity_restricted_guilds",
            "default_guilds_activity_restricted",
            "activity_joining_restricted_guilds",
            "message_request_restricted_guilds",
            "default_message_request_restricted",
        ):
            if field in kwargs:
                if field.endswith("_guilds"):
                    privacy[field.replace("_guilds", "_guild_ids")] = [
                        g.id for g in kwargs.pop(field)
                    ]
                else:
                    privacy[field] = kwargs.pop(field)

        debug = {}
        for field in ("rtc_panel_show_voice_states",):
            if field in kwargs:
                debug[field] = kwargs.pop(field)

        game_library = {}
        for field in (
            "install_shortcut_desktop",
            "install_shortcut_start_menu",
            "disable_games_tab",
        ):
            if field in kwargs:
                game_library[field] = kwargs.pop(field)

        status = {}
        if "status" in kwargs:
            status["status"] = _ocast(kwargs.pop("status"), str)
        if "custom_activity" in kwargs:
            status["custom_status"] = (
                kwargs.pop("custom_activity").to_settings_dict()
                if kwargs["custom_activity"] not in {MISSING, None}
                else MISSING
            )
        for field in ("show_current_game",):
            if field in kwargs:
                status[field] = kwargs.pop(field)

        localization = {}
        if "locale" in kwargs:
            localization["locale"] = _ocast(kwargs.pop("locale"), str)
        for field in ("timezone_offset",):
            if field in kwargs:
                localization[field] = kwargs.pop(field)

        appearance = {}
        if "theme" in kwargs:
            appearance["theme"] = _ocast(kwargs.pop("theme"), int)
        if "client_theme" in kwargs:
            provided: tuple = kwargs.pop("client_theme")
            client_theme_settings = {} if provided is not MISSING else MISSING
            if provided:
                if provided[0] is not MISSING:
                    client_theme_settings["primary_color"] = provided[0]
                if len(provided) > 1 and provided[1] is not MISSING:
                    client_theme_settings["background_gradient_preset_id"] = provided[1]
                if len(provided) > 2 and provided[2] is not MISSING:
                    client_theme_settings["background_gradient_angle"] = float(
                        provided[2]
                    )
                appearance["client_theme_settings"] = client_theme_settings
        if "disable_mobile_redesign" in kwargs:
            appearance["mobile_redesign_disabled"] = kwargs.pop(
                "disable_mobile_redesign"
            )
        for field in ("developer_mode",):
            if field in kwargs:
                appearance[field] = kwargs.pop(field)

        guild_folders = {}
        if "guild_folders" in kwargs:
            guild_folders["folders"] = (
                [f.to_dict() for f in kwargs.pop("guild_folders")]
                if kwargs["guild_folders"] is not MISSING
                else MISSING
            )
        if "guild_positions" in kwargs:
            guild_folders["guild_positions"] = (
                [g.id for g in kwargs.pop("guild_positions")]
                if kwargs["guild_positions"] is not MISSING
                else MISSING
            )

        audio_context_settings = {}
        if "user_audio_settings" in kwargs:
            audio_context_settings["user"] = (
                {s.id: s.to_dict() for s in kwargs.pop("user_audio_settings")}
                if kwargs["user_audio_settings"] is not MISSING
                else MISSING
            )
        if "stream_audio_settings" in kwargs:
            audio_context_settings["stream"] = (
                {s.id: s.to_dict() for s in kwargs.pop("stream_audio_settings")}
                if kwargs["stream_audio_settings"] is not MISSING
                else MISSING
            )

        communities = {}
        if "home_auto_navigation" in kwargs:
            communities["disable_home_auto_nav"] = (
                not kwargs.pop("home_auto_navigation")
                if kwargs["home_auto_navigation"] is not MISSING
                else MISSING
            )

        # Now, we do the actual patching
        existing = self.to_dict()
        payload = {}
        for subsetting in (
            "versions",
            "inbox",
            "guilds",
            "user_content",
            "voice_and_video",
            "text_and_images",
            "notifications",
            "privacy",
            "debug",
            "game_library",
            "status",
            "localization",
            "appearance",
            "guild_folders",
            "audio_context_settings",
            "communities",
        ):
            subsetting_dict = locals()[subsetting]
            if subsetting_dict:
                original = existing.get(subsetting, {})
                original.update(subsetting_dict)
                for k, v in dict(original).items():
                    if v is MISSING:
                        del original[k]
                payload[subsetting] = original

        state = self._state
        require_version = (
            self.data_version if require_version == True else require_version
        )
        ret = await state.http.edit_proto_settings(
            1, self.dict_to_base64(payload), require_version or None
        )
        return UserSettings(state, ret["settings"])


class GuildFolder:
    """Represents a guild folder.

    All properties have setters to faciliate editing the class for use with :meth:`UserSettings.edit`.

    .. container:: operations

        .. describe:: str(x)

            Returns the folder's name.

        .. describe:: len(x)

            Returns the number of guilds in the folder.

    .. versionadded:: 1.9

    .. versionchanged:: 2.0

        Removed various operations and made ``id`` and ``name`` optional.

    .. note::

        Guilds not in folders *are* actually in folders API wise, with them being the only member.

        These folders do not have an ID or name.

    Attributes
    ----------
    id: Optional[:class:`int`]
        The ID of the folder.
    name: Optional[:class:`str`]
        The name of the folder.
    """

    __slots__ = ("_state", "id", "name", "_colour", "_guild_ids")

    def __init__(
        self,
        *,
        id: Optional[int] = None,
        name: Optional[str] = None,
        colour: Optional[Colour] = None,
        guilds: Sequence[Snowflake] = MISSING,
    ):
        self._state: Optional[ConnectionState] = None
        self.id: Optional[int] = id
        self.name: Optional[str] = name
        self._colour: Optional[int] = colour.value if colour else None
        self._guild_ids: List[int] = [guild.id for guild in guilds] if guilds else []

    def __str__(self) -> str:
        return self.name or ", ".join(
            guild.name
            for guild in [guild for guild in self.guilds if isinstance(guild, Guild)]
        )

    def __repr__(self) -> str:
        return f"<GuildFolder id={self.id} name={self.name!r} guilds={self.guilds!r}>"

    def __len__(self) -> int:
        return len(self._guild_ids)

    @classmethod
    def _from_legacy_settings(
        cls, *, data: Dict[str, Any], state: ConnectionState
    ) -> Self:
        self = cls.__new__(cls)
        self._state = state
        self.id = _get_as_snowflake(data, "id")
        self.name = data.get("name")
        self._colour = data.get("color")
        self._guild_ids = [int(guild_id) for guild_id in data["guild_ids"]]
        return self

    @classmethod
    def _from_settings(cls, *, data: Any, state: ConnectionState) -> Self:
        """
        message GuildFolder {
            repeated fixed64 guild_ids = 1;
            optional google.protobuf.Int64Value id = 2;
            optional google.protobuf.StringValue name = 3;
            optional google.protobuf.UInt64Value color = 4;
        }
        """
        self = cls.__new__(cls)
        self._state = state
        self.id = data.id.value
        self.name = data.name.value
        self._colour = data.color.value if data.HasField("color") else None
        self._guild_ids = data.guild_ids
        return self

    def _get_guild(self, id, /) -> Union[Guild, Object]:
        from .guild import Guild  # circular import

        id = int(id)
        return (
            self._state._get_or_create_unavailable_guild(id)
            if self._state
            else Object(id=id, type=Guild)
        )

    def to_dict(self) -> dict:
        ret = {}
        if self.id is not None:
            ret["id"] = self.id
        if self.name is not None:
            ret["name"] = self.name
        if self._colour is not None:
            ret["color"] = self._colour
        ret["guild_ids"] = [str(guild_id) for guild_id in self._guild_ids]
        return ret

    def copy(self) -> Self:
        """Returns a shallow copy of the folder."""
        return self.__class__._from_legacy_settings(
            data=self.to_dict(), state=self._state
        )  # type: ignore

    def add_guild(self, guild: Snowflake) -> Self:
        """Adds a guild to the folder.

        This function returns the class instance to allow for fluent-style
        chaining.

        .. versionadded:: 2.0

        Parameters
        -----------
        guild: :class:`abc.Snowflake`
            The guild to add to the folder.
        """
        self._guild_ids.append(guild.id)
        return self

    def insert_guild_at(self, index: int, guild: Snowflake) -> Self:
        """Inserts a guild before a specified index to the folder.

        This function returns the class instance to allow for fluent-style
        chaining.

        .. versionadded:: 2.0

        Parameters
        -----------
        index: :class:`int`
            The index of where to insert the field.
        guild: :class:`abc.Snowflake`
            The guild to add to the folder.
        """
        self._guild_ids.insert(index, guild.id)
        return self

    def clear_guilds(self) -> None:
        """Removes all guilds from this folder.

        .. versionadded:: 2.0
        """
        self._guild_ids.clear()

    def remove_guild(self, index: int) -> None:
        """Removes a guild at a specified index.

        If the index is invalid or out of bounds then the error is
        silently swallowed.

        .. note::

            When deleting a field by index, the index of the other fields
            shift to fill the gap just like a regular list.

        .. versionadded:: 2.0

        Parameters
        -----------
        index: :class:`int`
            The index of the field to remove.
        """
        try:
            del self._guild_ids[index]
        except IndexError:
            pass

    def set_guild_at(self, index: int, guild: Snowflake) -> Self:
        """Modifies a guild to the guild object.

        The index must point to a valid pre-existing guild.

        This function returns the class instance to allow for fluent-style
        chaining.

        .. versionadded:: 2.0

        Parameters
        -----------
        index: :class:`int`
            The index of the field to modify.
        guild: :class:`abc.Snowflake`
            The guild to add to the folder.

        Raises
        -------
        IndexError
            An invalid index was provided.
        """
        self._guild_ids[index] = guild.id

        try:
            self._guild_ids[index] = guild.id
        except (TypeError, IndexError):
            raise IndexError("field index out of range")
        return self

    @property
    def guilds(self) -> List[Union[Guild, Object]]:
        """List[Union[:class:`Guild`, :class:`Object`]]: The guilds in the folder. Always :class:`Object` if state is not attached."""
        return [self._get_guild(guild_id) for guild_id in self._guild_ids]

    @guilds.setter
    def guilds(self, value: Sequence[Snowflake]) -> None:
        self._guild_ids = [guild.id for guild in value]

    @property
    def colour(self) -> Optional[Colour]:
        """Optional[:class:`Colour`]: The colour code of the folder. There is an alias for this named :attr:`colour`."""
        return Colour(self._colour) if self._colour is not None else None

    @colour.setter
    def colour(self, value: Optional[Union[int, Colour]]) -> None:
        if value is None:
            self._colour = None
        elif isinstance(value, Colour):
            self._colour = value.value
        elif isinstance(value, int):
            self._colour = value
        else:
            raise TypeError(
                f"Expected discord.Colour, int, or None but received {value.__class__.__name__} instead."
            )

    @property
    def color(self) -> Optional[Colour]:
        """Optional[:class:`Colour`]: The colour code of the folder. There is an alias for this named :attr:`colour`."""
        return self.colour

    @color.setter
    def color(self, value: Optional[Union[int, Colour]]) -> None:
        self.colour = value


class GuildProgress:
    """Represents a guild's settings revolving around upsells, promotions, and feature progress.

    All properties have setters to faciliate editing the class for use with :meth:`UserSettings.edit`.

    .. versionadded:: 2.0

    Attributes
    ----------
    guild_id: :class:`int`
        The ID of the guild.
    recents_dismissed_at: Optional[:class:`datetime.datetime`]
        When the guild recents were last dismissed.
    """

    __slots__ = (
        "guild_id",
        "_hub_progress",
        "_onboarding_progress",
        "recents_dismissed_at",
        "_dismissed_contents",
        "_collapsed_channel_ids",
        "_state",
    )

    def __init__(
        self,
        guild_id: int,
        *,
        hub_progress: HubProgressFlags,
        onboarding_progress: OnboardingProgressFlags,
        recents_dismissed_at: Optional[datetime] = None,
        dismissed_contents: Sequence[int] = MISSING,
        collapsed_channels: List[Snowflake] = MISSING,
    ) -> None:
        self._state: Optional[ConnectionState] = None
        self.guild_id = guild_id
        self._hub_progress = hub_progress.value
        self._onboarding_progress = onboarding_progress.value
        self.recents_dismissed_at: Optional[datetime] = recents_dismissed_at
        self._dismissed_contents = self._pack_dismissed_contents(
            dismissed_contents or []
        )
        self._collapsed_channel_ids = [
            channel.id for channel in collapsed_channels
        ] or []

    def __repr__(self) -> str:
        return f"<GuildProgress guild_id={self.guild_id} hub_progress={self.hub_progress!r} onboarding_progress={self.onboarding_progress!r}>"

    @classmethod
    def _from_settings(
        cls, guild_id: int, *, data: Any, state: ConnectionState
    ) -> Self:
        """
        message ChannelSettings {
            bool collapsed_in_inbox = 1;
        }

        message GuildSettings {
            map<fixed64, ChannelSettings> channels = 1;
            uint32 hub_progress = 2;
            uint32 guild_onboarding_progress = 3;
            optional google.protobuf.Timestamp guild_recents_dismissed_at = 4;
            bytes dismissed_guild_content = 5;
        }

        message AllGuildSettings {
            map<fixed64, GuildSettings> guilds = 1;
        }
        """
        self = cls.__new__(cls)
        self._state = state
        self.guild_id = guild_id
        self._hub_progress = data.hub_progress
        self._onboarding_progress = data.guild_onboarding_progress
        self.recents_dismissed_at = (
            data.guild_recents_dismissed_at.ToDatetime(tzinfo=timezone.utc)
            if data.HasField("guild_recents_dismissed_at")
            else None
        )
        self._dismissed_contents = data.dismissed_guild_content
        self._collapsed_channel_ids = [
            channel_id
            for channel_id, settings in data.channels.items()
            if settings.collapsed_in_inbox
        ]
        return self

    def _get_channel(self, id: int, /) -> Union[GuildChannel, Object]:
        id = int(id)
        return (
            self.guild.get_channel(id) or Object(id=id)
            if self.guild is not None
            else Object(id=id)
        )

    def to_dict(self) -> Dict[str, Any]:
        data = {
            "hub_progress": self._hub_progress,
            "guild_onboarding_progress": self._onboarding_progress,
            "dismissed_guild_content": self._dismissed_contents,
            "channels": {
                id: {"collapsed_in_inbox": True} for id in self._collapsed_channel_ids
            },
        }
        if self.recents_dismissed_at is not None:
            data["guild_recents_dismissed_at"] = self.recents_dismissed_at.isoformat()
        return data

    def copy(self) -> Self:
        """Returns a shallow copy of the progress settings."""
        cls = self.__class__(
            self.guild_id,
            hub_progress=self.hub_progress,
            onboarding_progress=self.onboarding_progress,
            recents_dismissed_at=self.recents_dismissed_at,
            dismissed_contents=self.dismissed_contents,
            collapsed_channels=self.collapsed_channels,
        )  # type: ignore
        cls._state = self._state
        return cls

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild this progress belongs to. ``None`` if state is not attached."""
        return (
            self._state._get_or_create_unavailable_guild(self.guild_id)
            if self._state is not None
            else None
        )

    @property
    def hub_progress(self) -> HubProgressFlags:
        """:class:`HubProgressFlags`: The hub's usage and feature progress."""
        return HubProgressFlags._from_value(self._hub_progress)

    @hub_progress.setter
    def hub_progress(self, value: HubProgressFlags) -> None:
        self._hub_progress = value.value

    @property
    def onboarding_progress(self) -> OnboardingProgressFlags:
        """:class:`OnboardingProgressFlags`: The guild's onboarding usage and feature progress."""
        return OnboardingProgressFlags._from_value(self._onboarding_progress)

    @onboarding_progress.setter
    def onboarding_progress(self, value: OnboardingProgressFlags) -> None:
        self._onboarding_progress = value.value

    @staticmethod
    def _pack_dismissed_contents(contents: Sequence[int]) -> bytes:
        return struct.pack(f">{len(contents)}B", *contents)

    @property
    def dismissed_contents(self) -> Tuple[int, ...]:
        """Tuple[:class:`int`]: A list of enum values representing per-guild dismissable content in the app.

        .. note::

            For now, this just returns the raw values without converting to a proper enum,
            as the enum values change too often to be viably maintained.
        """
        contents = self._dismissed_contents
        return struct.unpack(f">{len(contents)}B", contents)

    @dismissed_contents.setter
    def dismissed_contents(self, value: Sequence[int]) -> None:
        self._dismissed_contents = self._pack_dismissed_contents(value)

    @property
    def collapsed_channels(self) -> List[Union[GuildChannel, Object]]:
        """List[Union[:class:`abc.GuildChannel`, :class:`Object`]]: A list of guild channels that are collapsed in the inbox. Always :class:`Object` if state is not attached."""
        return list(map(self._get_channel, self._collapsed_channel_ids))

    @collapsed_channels.setter
    def collapsed_channels(self, value: Sequence[Snowflake]) -> None:
        self._collapsed_channel_ids = [channel.id for channel in value]


class AudioContext:
    """Represents saved audio settings for a user or stream.

    All properties have setters to faciliate editing the class for use with :meth:`UserSettings.edit`.

    .. versionadded:: 2.0

    Attributes
    ----------
    user_id: :class:`int`
        The ID of the user.
    muted: :class:`bool`
        Whether the user or stream is muted.
    volume: :class:`float`
        The volume of the user or stream (0-100).
    modified_at: :class:`datetime.datetime`
        The time the settings were last modified.
    """

    __slots__ = ("_state", "user_id", "muted", "volume", "modified_at")

    def __init__(self, user_id: int, *, muted: bool = False, volume: float) -> None:
        self._state: Optional[ConnectionState] = None
        self.user_id = user_id
        self.muted = muted
        self.volume = volume
        self.modified_at = utcnow()

    def __repr__(self) -> str:
        return f"<AudioContext user_id={self.user_id} muted={self.muted} volume={self.volume} modified_at={self.modified_at!r}>"

    @classmethod
    def _from_settings(cls, user_id: int, *, data: Any, state: ConnectionState) -> Self:
        """
        message AudioContextSetting {
            bool muted = 1;
            float volume = 2;
            fixed64 modified_at = 3;
        }
        """
        self = cls.__new__(cls)
        self._state = state
        self.user_id = user_id
        self.muted = data.muted
        self.volume = data.volume
        self.modified_at = parse_timestamp(data.modified_at)
        return self

    def to_dict(self) -> Dict[str, Any]:
        """Converts the object to a dictionary."""
        return {
            "user_id": self.user_id,
            "muted": self.muted,
            "volume": self.volume,
            "modified_at": self.modified_at.isoformat(),
        }

    def copy(self) -> Self:
        """Returns a shallow copy of the audio context."""
        cls = self.__class__(self.user_id, muted=self.muted, volume=self.volume)
        cls.modified_at = self.modified_at
        cls._state = self._state
        return cls

    @property
    def user(self) -> Optional[User]:
        """Optional[:class:`User`]: The user the settings are for. ``None`` if state is not attached."""
        return self._state.get_user(self.user_id) if self._state is not None else None


class LegacyUserSettings:
    """Represents the legacy Discord client settings.

    .. versionadded:: 1.9

    .. deprecated:: 2.0

    .. note::

        Discord has migrated user settings to a new protocol buffer format.
        While these legacy settings still exist, they are no longer sent to newer clients (so they will have to be fetched).

        The new settings are available in :class:`UserSettings`, and this class has been deprecated and renamed to :class:`LegacyUserSettings`.
        All options in this class are available in the new format, and changes are reflected in both.

    Attributes
    ----------
    afk_timeout: :class:`int`
        How long (in seconds) the user needs to be AFK until Discord
        sends push notifications to mobile devices (30-600).
    allow_accessibility_detection: :class:`bool`
        Whether to allow Discord to track screen reader usage.
    animate_emojis: :class:`bool`
        Whether to animate emojis in the chat.
    contact_sync_enabled: :class:`bool`
        Whether to enable the contact sync on Discord mobile.
    convert_emoticons: :class:`bool`
        Whether to automatically convert emoticons into emojis (e.g. :) -> 😃).
    default_guilds_restricted: :class:`bool`
        Whether to automatically disable DMs between you and
        members of new guilds you join.
    detect_platform_accounts: :class:`bool`
        Whether to automatically detect accounts from services
        like Steam and Blizzard when you open the Discord client.
    developer_mode: :class:`bool`
        Whether to enable developer mode.
    disable_games_tab: :class:`bool`
        Whether to disable the showing of the Games tab.
    enable_tts_command: :class:`bool`
        Whether to allow TTS messages to be played/sent.
    gif_auto_play: :class:`bool`
        Whether to automatically play GIFs that are in the chat.
    inline_attachment_media: :class:`bool`
        Whether to display attachments when they are uploaded in chat.
    inline_embed_media: :class:`bool`
        Whether to display videos and images from links posted in chat.
    message_display_compact: :class:`bool`
        Whether to use the compact Discord display mode.
    native_phone_integration_enabled: :class:`bool`
        Whether to enable the new Discord mobile phone number friend
        requesting features.
    render_embeds: :class:`bool`
        Whether to render embeds that are sent in the chat.
    render_reactions: :class:`bool`
        Whether to render reactions that are added to messages.
    show_current_game: :class:`bool`
        Whether to display the game that you are currently playing.
    stream_notifications_enabled: :class:`bool`
        Whether stream notifications for friends will be received.
    timezone_offset: :class:`int`
        The timezone offset from UTC to use (in minutes).
    view_nsfw_commands: :class:`bool`
        Whether to show NSFW application commands in DMs.

        .. versionadded:: 2.0
    view_nsfw_guilds: :class:`bool`
        Whether to show NSFW guilds on iOS.
    """

    if TYPE_CHECKING:  # Fuck me
        afk_timeout: int
        allow_accessibility_detection: bool
        animate_emojis: bool
        contact_sync_enabled: bool
        convert_emoticons: bool
        default_guilds_restricted: bool
        detect_platform_accounts: bool
        developer_mode: bool
        disable_games_tab: bool
        enable_tts_command: bool
        gif_auto_play: bool
        inline_attachment_media: bool
        inline_embed_media: bool
        message_display_compact: bool
        native_phone_integration_enabled: bool
        render_embeds: bool
        render_reactions: bool
        show_current_game: bool
        stream_notifications_enabled: bool
        timezone_offset: int
        view_nsfw_commands: bool
        view_nsfw_guilds: bool

    def __init__(self, *, data, state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return "<LegacyUserSettings>"

    def _get_guild(self, id: int, /) -> Guild:
        return self._state._get_or_create_unavailable_guild(int(id))

    def _update(self, data: Dict[str, Any]) -> None:
        RAW_VALUES = {
            "afk_timeout",
            "allow_accessibility_detection",
            "animate_emojis",
            "contact_sync_enabled",
            "convert_emoticons",
            "default_guilds_restricted",
            "detect_platform_accounts",
            "developer_mode",
            "disable_games_tab",
            "enable_tts_command",
            "gif_auto_play",
            "inline_attachment_media",
            "inline_embed_media",
            "message_display_compact",
            "native_phone_integration_enabled",
            "render_embeds",
            "render_reactions",
            "show_current_game",
            "stream_notifications_enabled",
            "timezone_offset",
            "view_nsfw_commands",
            "view_nsfw_guilds",
        }

        for key, value in data.items():
            if key in RAW_VALUES:
                setattr(self, key, value)
            else:
                setattr(self, "_" + key, value)

    async def edit(self, **kwargs) -> Self:
        """|coro|

        Edits the client user's settings.

        .. versionchanged:: 2.0
            The edit is no longer in-place, instead the newly edited settings are returned.

        .. deprecated:: 2.0

        Parameters
        ----------
        activity_restricted_guilds: List[:class:`~discord.abc.Snowflake`]
            A list of guilds that your current activity will not be shown in.

            .. versionadded:: 2.0
        activity_joining_restricted_guilds: List[:class:`~discord.abc.Snowflake`]
            A list of guilds that will not be able to join your current activity.

            .. versionadded:: 2.0
        afk_timeout: :class:`int`
            How long (in seconds) the user needs to be AFK until Discord
            sends push notifications to mobile device (30-600).
        allow_accessibility_detection: :class:`bool`
            Whether to allow Discord to track screen reader usage.
        animate_emojis: :class:`bool`
            Whether to animate emojis in the chat.
        animate_stickers: :class:`.StickerAnimationOptions`
            Whether to animate stickers in the chat.
        contact_sync_enabled: :class:`bool`
            Whether to enable the contact sync on Discord mobile.
        convert_emoticons: :class:`bool`
            Whether to automatically convert emoticons into emojis (e.g. :) -> 😃).
        default_guilds_restricted: :class:`bool`
            Whether to automatically disable DMs between you and
            members of new guilds you join.
        detect_platform_accounts: :class:`bool`
            Whether to automatically detect accounts from services
            like Steam and Blizzard when you open the Discord client.
        developer_mode: :class:`bool`
            Whether to enable developer mode.
        disable_games_tab: :class:`bool`
            Whether to disable the showing of the Games tab.
        enable_tts_command: :class:`bool`
            Whether to allow TTS messages to be played/sent.
        explicit_content_filter: :class:`.UserContentFilter`
            The filter for explicit content in all messages.
        friend_source_flags: :class:`.FriendSourceFlags`
            Who can add you as a friend.
        friend_discovery_flags: :class:`.FriendDiscoveryFlags`
            How you get recommended friends.
        gif_auto_play: :class:`bool`
            Whether to automatically play GIFs that are in the chat.
        guild_positions: List[:class:`~discord.abc.Snowflake`]
            A list of guilds in order of the guild/guild icons that are on
            the left hand side of the UI.
        inline_attachment_media: :class:`bool`
            Whether to display attachments when they are uploaded in chat.
        inline_embed_media: :class:`bool`
            Whether to display videos and images from links posted in chat.
        locale: :class:`.Locale`
            The :rfc:`3066` language identifier of the locale to use for the language
            of the Discord client.
        message_display_compact: :class:`bool`
            Whether to use the compact Discord display mode.
        native_phone_integration_enabled: :class:`bool`
            Whether to enable the new Discord mobile phone number friend
            requesting features.
        passwordless: :class:`bool`
            Whether to enable passwordless login.
        render_embeds: :class:`bool`
            Whether to render embeds that are sent in the chat.
        render_reactions: :class:`bool`
            Whether to render reactions that are added to messages.
        restricted_guilds: List[:class:`~discord.abc.Snowflake`]
            A list of guilds that you will not receive DMs from.
        show_current_game: :class:`bool`
            Whether to display the game that you are currently playing.
        stream_notifications_enabled: :class:`bool`
            Whether stream notifications for friends will be received.
        theme: :class:`.Theme`
            The overall theme of the Discord UI.
        timezone_offset: :class:`int`
            The timezone offset to use.
        view_nsfw_commands: :class:`bool`
            Whether to show NSFW application commands in DMs.

            .. versionadded:: 2.0
        view_nsfw_guilds: :class:`bool`
            Whether to show NSFW guilds on iOS.

            .. versionadded:: 2.0

        Raises
        -------
        HTTPException
            Editing the settings failed.

        Returns
        -------
        :class:`.UserSettings`
            The client user's updated settings.
        """
        return await self._state.client.edit_legacy_settings(**kwargs)

    @property
    def activity_restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that your current activity will not be shown in.

        .. versionadded:: 2.0
        """
        return list(
            map(self._get_guild, getattr(self, "_activity_restricted_guild_ids", []))
        )

    @property
    def activity_joining_restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that will not be able to join your current activity.

        .. versionadded:: 2.0
        """
        return list(
            map(
                self._get_guild,
                getattr(self, "_activity_joining_restricted_guild_ids", []),
            )
        )

    @property
    def animate_stickers(self) -> StickerAnimationOptions:
        """:class:`StickerAnimationOptions`: Whether to animate stickers in the chat."""
        return try_enum(StickerAnimationOptions, getattr(self, "_animate_stickers", 0))

    @property
    def custom_activity(self) -> Optional[CustomActivity]:
        """Optional[:class:`CustomActivity`]: The set custom activity."""
        return CustomActivity._from_legacy_settings(
            data=getattr(self, "_custom_status", None), state=self._state
        )

    @property
    def explicit_content_filter(self) -> UserContentFilter:
        """:class:`UserContentFilter`: The filter for explicit content in all messages."""
        return try_enum(UserContentFilter, getattr(self, "_explicit_content_filter", 0))

    @property
    def friend_source_flags(self) -> FriendSourceFlags:
        """:class:`FriendSourceFlags`: Who can add you as a friend."""
        return FriendSourceFlags._from_dict(
            getattr(self, "_friend_source_flags", {"all": True})
        )

    @property
    def friend_discovery_flags(self) -> FriendDiscoveryFlags:
        """:class:`FriendDiscoveryFlags`: How you get recommended friends."""
        return FriendDiscoveryFlags._from_value(
            getattr(self, "_friend_discovery_flags", 0)
        )

    @property
    def guild_folders(self) -> List[GuildFolder]:
        """List[:class:`GuildFolder`]: A list of guild folders."""
        state = self._state
        return [
            GuildFolder._from_legacy_settings(data=folder, state=state)
            for folder in getattr(self, "_guild_folders", [])
        ]

    @property
    def guild_positions(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds in order of the guild/guild icons that are on the left hand side of the UI."""
        return list(map(self._get_guild, getattr(self, "_guild_positions", [])))

    @property
    def locale(self) -> Locale:
        """:class:`Locale`: The :rfc:`3066` language identifier
        of the locale to use for the language of the Discord client.

        .. versionchanged:: 2.0
            This now returns a :class:`Locale` object instead of a string.
        """
        return try_enum(Locale, getattr(self, "_locale", "en-US"))

    @property
    def passwordless(self) -> bool:
        """:class:`bool`: Whether to enable passwordless login."""
        return getattr(self, "_passwordless", False)

    @property
    def restricted_guilds(self) -> List[Guild]:
        """List[:class:`Guild`]: A list of guilds that you will not receive DMs from."""
        return list(map(self._get_guild, getattr(self, "_restricted_guilds", [])))

    @property
    def status(self) -> Status:
        """Optional[:class:`Status`]: The configured status."""
        return try_enum(Status, getattr(self, "_status", "online"))

    @property
    def theme(self) -> Theme:
        """:class:`Theme`: The overall theme of the Discord UI."""
        return try_enum(Theme, getattr(self, "_theme", "dark"))  # Sane default :)


class MuteConfig:
    """An object representing an object's mute status.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two items are muted.

        .. describe:: x != y

            Checks if two items are not muted.

        .. describe:: str(x)

            Returns the mute status as a string.

        .. describe:: int(x)

            Returns the mute status as an int.

    Attributes
    ----------
    muted: :class:`bool`
        Indicates if the object is muted.
    until: Optional[:class:`datetime.datetime`]
        When the mute will expire.
    """

    def __init__(self, muted: bool, config: Dict[str, str]) -> None:
        until = parse_time(config.get("end_time"))
        if until is not None:
            if until <= utcnow():
                muted = False
                until = None

        self.muted: bool = muted
        self.until: Optional[datetime] = until

    def __repr__(self) -> str:
        return str(self.muted)

    def __int__(self) -> int:
        return int(self.muted)

    def __bool__(self) -> bool:
        return self.muted

    def __eq__(self, other: object) -> bool:
        return self.muted == bool(other)

    def __ne__(self, other: object) -> bool:
        return not self.muted == bool(other)


class ChannelSettings:
    """Represents a channel's notification settings.

    .. versionadded:: 2.0

    Attributes
    ----------
    level: :class:`NotificationLevel`
        The notification level for the channel.
    muted: :class:`MuteConfig`
        The mute configuration for the channel.
    collapsed: :class:`bool`
        Whether the channel is collapsed.
        Only applicable to channels of type :attr:`ChannelType.category`.
    """

    if TYPE_CHECKING:
        _channel_id: int
        level: NotificationLevel
        muted: MuteConfig
        collapsed: bool

    def __init__(
        self,
        guild_id: Optional[int] = None,
        *,
        data: Dict[str, Any],
        state: ConnectionState,
    ) -> None:
        self._guild_id = guild_id
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f"<ChannelSettings channel={self.channel} level={self.level} muted={self.muted} collapsed={self.collapsed}>"

    def _update(self, data: Dict[str, Any]) -> None:
        # We consider everything optional because this class can be constructed with no data
        # to represent the default settings
        self._channel_id = int(data["channel_id"])
        self.collapsed = data.get("collapsed", False)

        self.level = try_enum(NotificationLevel, data.get("message_notifications", 3))
        self.muted = MuteConfig(data.get("muted", False), data.get("mute_config") or {})

    @property
    def channel(self) -> Union[GuildChannel, PrivateChannel]:
        """Union[:class:`abc.GuildChannel`, :class:`abc.PrivateChannel`]: Returns the channel these settings are for."""
        guild = (
            self._state._get_or_create_unavailable_guild(self._guild_id)
            if self._guild_id
            else None
        )
        if guild:
            channel = guild.get_channel(self._channel_id)
        else:
            channel = self._state._get_private_channel(self._channel_id)
        if not channel:
            channel = Object(id=self._channel_id)
        return channel  # type: ignore # Lying for better developer UX

    async def edit(
        self,
        *,
        muted_until: Optional[Union[bool, datetime]] = MISSING,
        collapsed: bool = MISSING,
        level: NotificationLevel = MISSING,
    ) -> ChannelSettings:
        """|coro|

        Edits the channel's notification settings.

        All parameters are optional.

        Parameters
        -----------
        muted_until: Optional[Union[:class:`datetime.datetime`, :class:`bool`]]
            The date this channel's mute should expire.
            This can be ``True`` to mute indefinitely, or ``False``/``None`` to unmute.

            This must be a timezone-aware datetime object. Consider using :func:`utils.utcnow`.
        collapsed: :class:`bool`
            Indicates if the channel should be collapsed or not.
            Only applicable to channels of type :attr:`ChannelType.category`.
        level: :class:`NotificationLevel`
            Determines what level of notifications you receive for the channel.

        Raises
        -------
        HTTPException
            Editing the settings failed.

        Returns
        --------
        :class:`ChannelSettings`
            The new notification settings.
        """
        state = self._state
        guild_id = self._guild_id
        channel_id = self._channel_id
        payload = {}

        if muted_until is not MISSING:
            if not muted_until:
                payload["muted"] = False
            else:
                payload["muted"] = True
                if muted_until is True:
                    payload["mute_config"] = {
                        "selected_time_window": -1,
                        "end_time": None,
                    }
                else:
                    if muted_until.tzinfo is None:
                        raise TypeError(
                            "muted_until must be an aware datetime. Consider using discord.utils.utcnow() or datetime.datetime.now().astimezone() for local time."
                        )

                    mute_config = {
                        "selected_time_window": (
                            muted_until - utcnow()
                        ).total_seconds(),
                        "end_time": muted_until.isoformat(),
                    }
                    payload["mute_config"] = mute_config

        if collapsed is not MISSING:
            payload["collapsed"] = collapsed

        if level is not MISSING:
            payload["message_notifications"] = level.value

        fields = {"channel_overrides": {str(channel_id): payload}}
        data = await state.http.edit_guild_settings(guild_id or "@me", fields)

        override = find(
            lambda x: x.get("channel_id") == str(channel_id), data["channel_overrides"]
        ) or {"channel_id": channel_id}
        return ChannelSettings(guild_id, data=override, state=state)


class GuildSettings:
    """Represents a guild's notification settings.

    .. versionadded:: 2.0

    Attributes
    ----------
    level: :class:`NotificationLevel`
        The notification level for the guild.
    muted: :class:`MuteConfig`
        The mute configuration for the guild.
    suppress_everyone: :class:`bool`
        Whether to suppress @everyone/@here notifications.
    suppress_roles: :class:`bool`
        Whether to suppress role notifications.
    hide_muted_channels: :class:`bool`
        Whether to hide muted channels.
    mobile_push: :class:`bool`
        Whether to enable mobile push notifications.
    mute_scheduled_events: :class:`bool`
        Whether to mute scheduled events.
    notify_highlights: :class:`HighlightLevel`
        Whether to include highlights in notifications.
    version: :class:`int`
        The version of the guild's settings.
    """

    if TYPE_CHECKING:
        _channel_overrides: Dict[int, ChannelSettings]
        _guild_id: Optional[int]
        level: NotificationLevel
        muted: MuteConfig
        suppress_everyone: bool
        suppress_roles: bool
        hide_muted_channels: bool
        mobile_push: bool
        mute_scheduled_events: bool
        notify_highlights: HighlightLevel
        version: int

    def __init__(self, *, data: Dict[str, Any], state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f"<GuildSettings guild={self.guild!r} level={self.level} muted={self.muted} suppress_everyone={self.suppress_everyone} suppress_roles={self.suppress_roles}>"

    def _update(self, data: Dict[str, Any]) -> None:
        # We consider everything optional because this class can be constructed with no data
        # to represent the default settings
        self._guild_id = guild_id = _get_as_snowflake(data, "guild_id")
        self.level = try_enum(NotificationLevel, data.get("message_notifications", 3))
        self.suppress_everyone = data.get("suppress_everyone", False)
        self.suppress_roles = data.get("suppress_roles", False)
        self.hide_muted_channels = data.get("hide_muted_channels", False)
        self.mobile_push = data.get("mobile_push", True)
        self.mute_scheduled_events = data.get("mute_scheduled_events", False)
        self.notify_highlights = try_enum(
            HighlightLevel, data.get("notify_highlights", 0)
        )
        self.version = data.get("version", -1)  # Overriden by real data

        self.muted = MuteConfig(data.get("muted", False), data.get("mute_config") or {})
        self._channel_overrides = overrides = {}
        state = self._state
        for override in data.get("channel_overrides", []):
            channel_id = int(override["channel_id"])
            overrides[channel_id] = ChannelSettings(
                guild_id, data=override, state=state
            )

    @property
    def guild(self) -> Union[Guild, ClientUser]:
        """Union[:class:`Guild`, :class:`ClientUser`]: Returns the guild that these settings are for.

        If the returned value is a :class:`ClientUser` then the settings are for the user's private channels.
        """
        if self._guild_id:
            return self._state._get_or_create_unavailable_guild(self._guild_id)
        return self._state.user  # type: ignore # Should always be present here

    @property
    def channel_overrides(self) -> List[ChannelSettings]:
        """List[:class:`ChannelSettings`: Returns a list of all the overrided channel notification settings."""
        return list(self._channel_overrides.values())

    async def edit(
        self,
        muted_until: Optional[Union[bool, datetime]] = MISSING,
        level: NotificationLevel = MISSING,
        suppress_everyone: bool = MISSING,
        suppress_roles: bool = MISSING,
        mobile_push: bool = MISSING,
        hide_muted_channels: bool = MISSING,
        mute_scheduled_events: bool = MISSING,
        notify_highlights: HighlightLevel = MISSING,
    ) -> Optional[GuildSettings]:
        """|coro|

        Edits the guild's notification settings.

        All parameters are optional.

        Parameters
        -----------
        muted_until: Optional[Union[:class:`datetime.datetime`, :class:`bool`]]
            The date this guild's mute should expire.
            This can be ``True`` to mute indefinitely, or ``False``/``None`` to unmute.

            This must be a timezone-aware datetime object. Consider using :func:`utils.utcnow`.
        level: :class:`NotificationLevel`
            Determines what level of notifications you receive for the guild.
        suppress_everyone: :class:`bool`
            Indicates if @everyone mentions should be suppressed for the guild.
        suppress_roles: :class:`bool`
            Indicates if role mentions should be suppressed for the guild.
        mobile_push: :class:`bool`
            Indicates if push notifications should be sent to mobile devices for this guild.
        hide_muted_channels: :class:`bool`
            Indicates if channels that are muted should be hidden from the sidebar.
        mute_scheduled_events: :class:`bool`
            Indicates if scheduled events should be muted.
        notify_highlights: :class:`HighlightLevel`
            Indicates if highlights should be included in notifications.

        Raises
        -------
        HTTPException
            Editing the settings failed.

        Returns
        --------
        :class:`GuildSettings`
            The new notification settings.
        """
        payload = {}

        if muted_until is not MISSING:
            if not muted_until:
                payload["muted"] = False
            else:
                payload["muted"] = True
                if muted_until is True:
                    payload["mute_config"] = {
                        "selected_time_window": -1,
                        "end_time": None,
                    }
                else:
                    if muted_until.tzinfo is None:
                        raise TypeError(
                            "muted_until must be an aware datetime. Consider using discord.utils.utcnow() or datetime.datetime.now().astimezone() for local time."
                        )

                    mute_config = {
                        "selected_time_window": (
                            muted_until - utcnow()
                        ).total_seconds(),
                        "end_time": muted_until.isoformat(),
                    }
                    payload["mute_config"] = mute_config

        if level is not MISSING:
            payload["message_notifications"] = level.value

        if suppress_everyone is not MISSING:
            payload["suppress_everyone"] = suppress_everyone

        if suppress_roles is not MISSING:
            payload["suppress_roles"] = suppress_roles

        if mobile_push is not MISSING:
            payload["mobile_push"] = mobile_push

        if hide_muted_channels is not MISSING:
            payload["hide_muted_channels"] = hide_muted_channels

        if mute_scheduled_events is not MISSING:
            payload["mute_scheduled_events"] = mute_scheduled_events

        if notify_highlights is not MISSING:
            payload["notify_highlights"] = notify_highlights.value

        data = await self._state.http.edit_guild_settings(
            self._guild_id or "@me", payload
        )
        return GuildSettings(data=data, state=self._state)


class TrackingSettings:
    """Represents your Discord tracking settings.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: bool(x)

            Checks if any tracking settings are enabled.

    Attributes
    ----------
    personalization: :class:`bool`
        Whether you have consented to your data being used for personalization.
    usage_statistics: :class:`bool`
        Whether you have consented to your data being used for usage statistics.
    """

    __slots__ = ("_state", "personalization", "usage_statistics")

    def __init__(
        self, *, data: Dict[str, Dict[str, bool]], state: ConnectionState
    ) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f"<TrackingSettings personalization={self.personalization} usage_statistics={self.usage_statistics}>"

    def __bool__(self) -> bool:
        return any({self.personalization, self.usage_statistics})

    def _update(self, data: Dict[str, Dict[str, bool]]):
        self.personalization = data.get("personalization", {}).get("consented", False)
        self.usage_statistics = data.get("usage_statistics", {}).get("consented", False)

    @overload
    async def edit(self) -> None: ...

    @overload
    async def edit(
        self,
        *,
        personalization: bool = ...,
        usage_statistics: bool = ...,
    ) -> None: ...

    async def edit(self, **kwargs) -> None:
        """|coro|

        Edits your tracking settings.

        Parameters
        ----------
        personalization: :class:`bool`
            Whether you have consented to your data being used for personalization.
        usage_statistics: :class:`bool`
            Whether you have consented to your data being used for usage statistics.
        """
        payload = {
            "grant": [k for k, v in kwargs.items() if v is True],
            "revoke": [k for k, v in kwargs.items() if v is False],
        }
        data = await self._state.http.edit_tracking(payload)
        self._update(data)


class EmailSettings:
    """Represents email communication preferences.

    .. versionadded:: 2.0

    Attributes
    ----------
    initialized: :class:`bool`
        Whether the email communication preferences have been initialized.
    communication: :class:`bool`
        Whether you want to receive emails for missed calls/messages.
    social: :class:`bool`
        Whether you want to receive emails for friend requests/suggestions or events.
    recommendations_and_events: :class:`bool`
        Whether you want to receive emails for recommended servers and events.
    tips: :class:`bool`
        Whether you want to receive emails for advice and tricks.
    updates_and_announcements: :class:`bool`
        Whether you want to receive emails for updates and new features.
    """

    __slots__ = (
        "_state",
        "initialized",
        "communication",
        "social",
        "recommendations_and_events",
        "tips",
        "updates_and_announcements",
    )

    def __init__(self, *, data: dict, state: ConnectionState):
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f"<EmailSettings initialized={self.initialized}>"

    def _update(self, data: dict):
        self.initialized = data.get("initialized", False)
        categories = data.get("categories", {})
        self.communication = categories.get("communication", False)
        self.social = categories.get("social", False)
        self.recommendations_and_events = categories.get(
            "recommendations_and_events", False
        )
        self.tips = categories.get("tips", False)
        self.updates_and_announcements = categories.get(
            "updates_and_announcements", False
        )

    @overload
    async def edit(self) -> None: ...

    @overload
    async def edit(
        self,
        *,
        communication: bool = MISSING,
        social: bool = MISSING,
        recommendations_and_events: bool = MISSING,
        tips: bool = MISSING,
        updates_and_announcements: bool = MISSING,
    ) -> None: ...

    async def edit(self, **kwargs) -> None:
        """|coro|

        Edits the email settings.

        All parameters are optional.

        Parameters
        -----------
        communication: :class:`bool`
            Indicates if you want to receive communication emails.
        social: :class:`bool`
            Indicates if you want to receive social emails.
        recommendations_and_events: :class:`bool`
            Indicates if you want to receive recommendations and events emails.
        tips: :class:`bool`
            Indicates if you want to receive tips emails.
        updates_and_announcements: :class:`bool`
            Indicates if you want to receive updates and announcements emails.

        Raises
        -------
        HTTPException
            Editing the settings failed.
        """
        payload = {}

        # It seems that initialized is settable, but it doesn't do anything
        # So we support just in case but leave it undocumented
        initialized = kwargs.pop("initialized", None)
        if initialized is not None:
            payload["initialized"] = initialized
        if kwargs:
            payload["categories"] = kwargs

        data = await self._state.http.edit_email_settings(**payload)
        self._update(data)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from .enums import ExpireBehaviour, try_enum
from .user import User
from .utils import MISSING, _get_as_snowflake, parse_time, utcnow

__all__ = (
    "IntegrationAccount",
    "Integration",
    "StreamIntegration",
    "BotIntegration",
)

if TYPE_CHECKING:
    from datetime import datetime

    from .application import IntegrationApplication
    from .guild import Guild
    from .role import Role
    from .state import ConnectionState
    from .types.integration import (
        BotIntegration as BotIntegrationPayload,
    )
    from .types.integration import (
        Integration as IntegrationPayload,
    )
    from .types.integration import (
        IntegrationAccount as IntegrationAccountPayload,
    )
    from .types.integration import (
        IntegrationType,
    )
    from .types.integration import (
        StreamIntegration as StreamIntegrationPayload,
    )


class IntegrationAccount:
    """Represents an integration account.

    .. versionadded:: 1.4

    Attributes
    -----------
    id: :class:`str`
        The account ID.
    name: :class:`str`
        The account name.
    """

    __slots__ = ("id", "name")

    def __init__(self, data: IntegrationAccountPayload) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]

    def __repr__(self) -> str:
        return f"<IntegrationAccount id={self.id} name={self.name!r}>"


class Integration:
    """Represents a guild integration.

    .. versionadded:: 1.4

    Attributes
    -----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type.
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    account: :class:`IntegrationAccount`
        The account linked to this integration.
    user: Optional[:class:`User`]
        The user that added this integration, if available.
    """

    __slots__ = (
        "guild",
        "id",
        "_state",
        "type",
        "name",
        "account",
        "user",
        "enabled",
    )

    def __init__(self, *, data: IntegrationPayload, guild: Guild) -> None:
        self.guild: Guild = guild
        self._state: ConnectionState = guild._state
        self._from_data(data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r}>"

    def _from_data(self, data: IntegrationPayload) -> None:
        self.id: int = int(data["id"])
        self.type: IntegrationType = data["type"]
        self.name: str = data["name"]
        self.account: IntegrationAccount = IntegrationAccount(data["account"])

        user = data.get("user")
        self.user: Optional[User] = User(state=self._state, data=user) if user else None
        self.enabled: bool = data.get("enabled", True)

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Deletes the integration.

        You must have :attr:`~Permissions.manage_guild` to do this.

        Parameters
        -----------
        reason: :class:`str`
            The reason the integration was deleted. Shows up on the audit log.

            .. versionadded:: 2.0

        Raises
        -------
        Forbidden
            You do not have permission to delete the integration.
        HTTPException
            Deleting the integration failed.
        """
        await self._state.http.delete_integration(self.guild.id, self.id, reason=reason)
        self.enabled = False


class StreamIntegration(Integration):
    """Represents a stream integration for Twitch or YouTube.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. Twitch).
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    syncing: :class:`bool`
        Where the integration is currently syncing.
    enable_emoticons: :class:`bool`
        Whether emoticons should be synced for this integration (currently twitch only).
    expire_behaviour: :class:`ExpireBehaviour`
        The behaviour of expiring subscribers. Aliased to ``expire_behavior`` as well.
    expire_grace_period: :class:`int`
        The grace period (in days) for expiring subscribers.
    user: Optional[:class:`User`]
        The user for the integration, if available.
    account: :class:`IntegrationAccount`
        The integration account information.
    synced_at: :class:`datetime.datetime`
        An aware UTC datetime representing when the integration was last synced.
    """

    __slots__ = (
        "revoked",
        "expire_behaviour",
        "expire_grace_period",
        "synced_at",
        "_role_id",
        "syncing",
        "enable_emoticons",
        "subscriber_count",
    )

    def _from_data(self, data: StreamIntegrationPayload) -> None:
        super()._from_data(data)
        self.revoked: bool = data.get("revoked", False)
        self.expire_behaviour: ExpireBehaviour = try_enum(
            ExpireBehaviour, data.get("expire_behaviour", 0)
        )
        self.expire_grace_period: int = data.get("expire_grace_period", 1)
        self.synced_at: datetime = (
            parse_time(data["synced_at"]) if "synced_at" in data else utcnow()
        )
        self._role_id: Optional[int] = _get_as_snowflake(data, "role_id")
        self.syncing: bool = data.get("syncing", False)
        self.enable_emoticons: bool = data.get("enable_emoticons", True)
        self.subscriber_count: int = data.get("subscriber_count", 0)

    @property
    def expire_behavior(self) -> ExpireBehaviour:
        """:class:`ExpireBehaviour`: An alias for :attr:`expire_behaviour`."""
        return self.expire_behaviour

    @property
    def role(self) -> Optional[Role]:
        """Optional[:class:`Role`] The role which the integration uses for subscribers."""
        # The key is `int` but `int | None` will return `None` anyway.
        return self.guild.get_role(self._role_id)  # type: ignore

    async def edit(
        self,
        *,
        expire_behaviour: ExpireBehaviour = MISSING,
        expire_grace_period: int = MISSING,
        enable_emoticons: bool = MISSING,
    ) -> None:
        """|coro|

        Edits the integration.

        You must have :attr:`~Permissions.manage_guild` to do this.

        Parameters
        -----------
        expire_behaviour: :class:`ExpireBehaviour`
            The behaviour when an integration subscription lapses. Aliased to ``expire_behavior`` as well.
        expire_grace_period: :class:`int`
            The period (in days) where the integration will ignore lapsed subscriptions.
        enable_emoticons: :class:`bool`
            Where emoticons should be synced for this integration (currently twitch only).

        Raises
        -------
        Forbidden
            You do not have permission to edit the integration.
        HTTPException
            Editing the guild failed.
        """
        payload: Dict[str, Any] = {}
        if expire_behaviour is not MISSING:
            payload["expire_behavior"] = int(expire_behaviour)

        if expire_grace_period is not MISSING:
            payload["expire_grace_period"] = expire_grace_period

        if enable_emoticons is not MISSING:
            payload["enable_emoticons"] = enable_emoticons

        await self._state.http.edit_integration(self.guild.id, self.id, **payload)

    async def sync(self) -> None:
        """|coro|

        Syncs the integration.

        You must have :attr:`~Permissions.manage_guild` to do this.

        Raises
        -------
        Forbidden
            You do not have permission to sync the integration.
        HTTPException
            Syncing the integration failed.
        """
        await self._state.http.sync_integration(self.guild.id, self.id)
        self.synced_at = utcnow()

    async def disable(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Disables the integration.

        This is an alias of :meth:`Integration.delete`.

        You must have :attr:`~Permissions.manage_guild` to do this.

        Parameters
        -----------
        reason: :class:`str`
            The reason the integration was disabled. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have permission to disable the integration.
        HTTPException
            Disabling the integration failed.
        """
        await self.delete(reason=reason)

    async def enable(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Enables the integration.

        You must have :attr:`~Permissions.manage_guild` to do this.

        Parameters
        -----------
        reason: :class:`str`
            The reason the integration was enabled. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have permission to enable the integration.
        HTTPException
            Enabling the integration failed.
        """
        await self._state.http.create_integration(
            self.guild.id, self.type, self.id, reason=reason
        )
        self.enabled = True


class BotIntegration(Integration):
    """Represents a bot integration on discord.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. Discord).
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    user: Optional[:class:`User`]
        The user that added this integration, if available.
    account: :class:`IntegrationAccount`
        The integration account information.
    application_id: :class:`int`
        The application ID of the integration.
    application: Optional[:class:`IntegrationApplication`]
        The application tied to this integration. Not available in some contexts.
    scopes: List[:class:`str`]
        The scopes the integration is authorized for.
    """

    __slots__ = ("application", "application_id", "scopes")

    def _from_data(self, data: BotIntegrationPayload) -> None:
        super()._from_data(data)
        self.application: Optional[IntegrationApplication] = (
            self._state.create_integration_application(data["application"])
            if "application" in data
            else None
        )
        self.application_id = (
            self.application.id if self.application else int(data["application_id"])
        )  # type: ignore # One or the other
        self.scopes: List[str] = data.get("scopes", [])


def _integration_factory(value: str) -> Tuple[Type[Integration], str]:
    if value == "discord":
        return BotIntegration, value
    elif value in ("twitch", "youtube"):
        return StreamIntegration, value
    else:
        return Integration, value



from discord_self._vendor.discord.types.audit_log import (
    AuditEntryInfo,
    AuditLog,
    AuditLogChange,
    AuditLogEntry,
    AuditLogEvent,
)
from discord_self._vendor.discord.types.channel import (
    ChannelType,
    DefaultReaction,
    ForumTag,
    PermissionOverwrite,
    PrivacyLevel,
    VideoQualityMode,
)
from discord_self._vendor.discord.types.guild import (
    DefaultMessageNotificationLevel,
    ExplicitContentFilterLevel,
    MFALevel,
    VerificationLevel,
)
from discord_self._vendor.discord.types.integration import (
    IntegrationExpireBehavior,
    PartialIntegration,
)
from discord_self._vendor.discord.types.role import Role
from discord_self._vendor.discord.types.scheduled_event import (
    EntityType,
    EventStatus,
    GuildScheduledEvent,
)
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.threads import Thread
from discord_self._vendor.discord.types.user import User
from discord_self._vendor.discord.types.webhook import Webhook

__all__ = [
    "AuditEntryInfo",
    "AuditLog",
    "AuditLogChange",
    "AuditLogEntry",
    "AuditLogEvent",
    "ChannelType",
    "DefaultMessageNotificationLevel",
    "DefaultReaction",
    "EntityType",
    "EventStatus",
    "ExplicitContentFilterLevel",
    "ForumTag",
    "GuildScheduledEvent",
    "IntegrationExpireBehavior",
    "MFALevel",
    "PartialIntegration",
    "PermissionOverwrite",
    "PrivacyLevel",
    "Role",
    "Snowflake",
    "Thread",
    "User",
    "VerificationLevel",
    "VideoQualityMode",
    "Webhook",
]



from setuptools import setup, find_packages

setup(
    name="discord_self",
    version="2.0.1.post3",
    description="A Python wrapper for the Discord user API",
    long_description="discord_self is a fork of discord.py-self that lives under the `discord_self` namespace in order to allow both selfbots and bots to run together.\n\n\ndiscord.py-self\n================\n\n.. image:: https://img.shields.io/endpoint?url=https%3A%2F%2Frunkit.io%2Fdamiankrawczyk%2Ftelegram-badge%2Fbranches%2Fmaster%3Furl%3Dhttps%3A%2F%2Ft.me%2Fdpy_self\n   :target: https://t.me/dpy_self\n   :alt: Telegram chat\n.. image:: https://img.shields.io/pypi/v/discord.py-self.svg\n   :target: https://pypi.python.org/pypi/discord.py-self\n   :alt: PyPI version info\n.. image:: https://img.shields.io/pypi/pyversions/discord.py.svg\n   :target: https://pypi.python.org/pypi/discord.py-self\n   :alt: PyPI supported Python versions\n.. image:: https://img.shields.io/pypi/dm/discord.py-self.svg\n   :target: https://pypi.python.org/pypi/discord.py-self\n   :alt: PyPI downloads per month\n\nA modern, easy to use, feature-rich, and async ready API wrapper for Discord's user API written in Python.\n\n| **Note:**\n| Automating user accounts is against the Discord ToS. This library is a proof of concept and I cannot recommend using it. Do so at your own risk.\n\nFork Changes\n------------\n\nThese changes have become too numerous to mention, so check out our `docs <https://discordpy-self.readthedocs.io/en/latest/index.html>`_.\n\n**Credits:**\n\n- `Rapptz <https://github.com/Rapptz>`_ for the original library this fork is based on. Without it, the project would not exist.\n- `arandomnewaccount <https://www.reddit.com/user/obviouslymymain123/>`_ for help when the project was first started.\n\nKey Features\n-------------\n\n- Modern Pythonic API using ``async`` and ``await``.\n- Proper rate limit handling.\n- Optimised in both speed and memory.\n- Mostly compatible with the upstream ``discord.py``.\n- Prevents user account automation detection.\n- Implements vast amounts of the user account-specific API. For a non-exhaustive list:\n\n  * Sessions\n  * Connections\n  * Relationships\n  * Protobuf user settings\n  * Application/team management\n  * Store/SKUs/entitlements\n  * Billing (e.g. subscriptions, payments, boosts, promotions, etc.)\n  * Interactions (slash commands, buttons, etc.)\n\nInstalling\n----------\n\n**Python 3.8 or higher is required.**\n\nTo install the library without full voice support, you can just run the following command:\n\n.. code:: sh\n\n    # Linux/macOS\n    python3 -m pip install -U discord.py-self\n\n    # Windows\n    py -3 -m pip install -U discord.py-self\n\nOtherwise to get voice support you should run the following command:\n\n.. code:: sh\n\n    # Linux/macOS\n    python3 -m pip install -U \"discord.py-self[voice]\"\n\n    # Windows\n    py -3 -m pip install -U discord.py-self[voice]\n\n\nTo install the development version, do the following:\n\n.. code:: sh\n\n    $ git clone https://github.com/dolfies/discord.py-self\n    $ cd discord.py-self\n    $ python3 -m pip install -U .[voice]\n\n\nOptional Packages\n~~~~~~~~~~~~~~~~~~\n\n* `PyNaCl <https://pypi.org/project/PyNaCl/>`__ (for voice support)\n\nPlease note that on Linux installing voice you must install the following packages via your favourite package manager (e.g. ``apt``, ``dnf``, etc) before running the above commands:\n\n* libffi-dev (or ``libffi-devel`` on some systems)\n* python-dev (e.g. ``python3.6-dev`` for Python 3.6)\n\nUsing with Upstream\n~~~~~~~~~~~~~~~~~~~~\n\nIf you would like to use the library alongside upstream ``discord.py``, you can install ``selfcord.py`` instead of ``discord.py-self``. Check out the `renamed branch <https://github.com/dolfies/discord.py-self/tree/renamed>`_ for more information.\n\nQuick Example\n--------------\n\n.. code:: py\n\n    import discord\n\n    class MyClient(discord.Client):\n        async def on_ready(self):\n            print('Logged on as', self.user)\n\n        async def on_message(self, message):\n            # only respond to ourselves\n            if message.author != self.user:\n                return\n\n            if message.content == 'ping':\n                await message.channel.send('pong')\n\n    client = MyClient()\n    client.run('token')\n\nBot Example\n~~~~~~~~~~~~~\n\n.. code:: py\n\n    import discord\n    from discord.ext import commands\n\n    bot = commands.Bot(command_prefix='>', self_bot=True)\n\n    @bot.command()\n    async def ping(ctx):\n        await ctx.send('pong')\n\n    bot.run('token')\n\nYou can find more examples in the examples directory.\n\nLinks\n------\n\n- `Documentation <https://discordpy-self.readthedocs.io/en/latest/index.html>`_\n- `Project updates <https://t.me/dpy_self>`_\n- `Discussion & support <https://t.me/dpy_self_discussions>`_\n\n\n<!-- vendorization-metadata: {commit_hash: 759ad60d8c47c375d4ddaf101b65b5210e813db0, upstream_version: 2.0.1, vendorized_version: 2.0.1.post3} -->",
    long_description_content_type="text/x-rst",
    author="Steven Van Ingelgem",
    author_email="steven@vaningelgem.be",
    url="https://github.com/svaningelgem/discord_self",
    license="MIT",
    python_requires=">=3.8.0",
    license_files=[
        "LICENSE",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    install_requires=[
        "aiohttp<4,>=3.7.4",
        "discord-protos<1.0.0",
    ],
    extras_require={
        "docs": [
            "sphinx==4.4.0",
            "sphinxcontrib-trio==1.1.2",
            "sphinxcontrib-websupport",
            "typing-extensions<5,>=4.3",
        ],
        "speed": [
            "orjson>=3.5.4",
            "aiodns>=1.1",
            "Brotli",
            "cchardet==2.1.7; python_version < \"3.10\"",
        ],
        "test": [
            "coverage[toml]",
            "pytest",
            "pytest-asyncio",
            "pytest-cov",
            "pytest-mock",
            "typing-extensions<5,>=4.3",
        ],
        "voice": [
            "PyNaCl<1.6,>=1.3.0",
        ],
    },
    project_urls={
        "Documentation": "https://discordpy-self.readthedocs.io/en/latest/",
        "Issue tracker": "https://github.com/dolfies/discord.py-self/issues",
        "Project updates": "https://t.me/dpy_self",
        "Discussion & support": "https://t.me/dpy_self_discussions",
    },
    packages=find_packages(),
)


"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union

from .. import discord
from ..discord import abc as ___vendorize__0
from .asset import Asset
from .colour import Colour
from .enums import (
    HypeSquadHouse,
    Locale,
    PremiumType,
    RelationshipAction,
    RelationshipType,
    try_enum,
)
from .errors import ClientException, NotFound
from .flags import PremiumUsageFlags, PrivateUserFlags, PublicUserFlags, PurchasedFlags
from .relationship import Relationship
from .utils import (
    MISSING,
    _bytes_to_base64_data,
    _get_as_snowflake,
    cached_slot_property,
    copy_doc,
    snowflake_time,
)
from .voice_client import VoiceClient

if TYPE_CHECKING:
    from datetime import datetime

    from typing_extensions import Self

    from .abc import T as ConnectReturn
    from .abc import VocalChannel
    from .calls import PrivateCall
    from .channel import DMChannel
    from .client import Client
    from .member import VoiceState
    from .message import Message
    from .profile import UserProfile
    from .state import ConnectionState
    from .types.channel import DMChannel as DMChannelPayload
    from .types.snowflake import Snowflake
    from .types.user import (
        PartialUser as PartialUserPayload,
    )
    from .types.user import (
        User as UserPayload,
    )


__all__ = (
    "User",
    "ClientUser",
    "Note",
)


class Note:
    """Represents a Discord note.

    .. container:: operations

        .. describe:: x == y
            Checks if two notes are equal.

        .. describe:: x != y
            Checks if two notes are not equal.

        .. describe:: hash(x)
            Returns the note's hash.

        .. describe:: str(x)
            Returns the note's content.
            Raises :exc:`ClientException` if the note is not fetched.

        .. describe:: bool(x)
            Returns the note's content as a boolean.

        .. describe:: len(x)
            Returns the note's length.

    .. versionadded:: 1.9

    Attributes
    -----------
    user_id: :class:`int`
        The user ID the note is for.
    """

    __slots__ = ("_state", "_value", "user_id", "_user")

    def __init__(
        self,
        state: ConnectionState,
        user_id: int,
        *,
        user: Optional[User] = None,
        note: Optional[str] = MISSING,
    ) -> None:
        self._state = state
        self._value: Optional[str] = note
        self.user_id: int = user_id
        self._user: Optional[User] = user

    @property
    def note(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the note.

        There is an alias for this called :attr:`value`.

        Raises
        -------
        ClientException
            Attempted to access note without fetching it.
        """
        if self._value is MISSING:
            raise ClientException("Note is not fetched")
        return self._value

    @property
    def value(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the note.

        This is an alias of :attr:`note`.

        Raises
        -------
        ClientException
            Attempted to access note without fetching it.
        """
        return self.note

    @property
    def user(self) -> Optional[User]:
        """Optional[:class:`User`]: Returns the user the note belongs to."""
        return self._state.get_user(self.user_id) or self._user

    async def fetch(self) -> Optional[str]:
        """|coro|

        Retrieves the note.

        Raises
        -------
        HTTPException
            Fetching the note failed.

        Returns
        --------
        Optional[:class:`str`]
            The note or ``None`` if it doesn't exist.
        """
        try:
            data = await self._state.http.get_note(self.user_id)
            self._value = data["note"]
            return data["note"]
        except NotFound:
            # A 404 means the note doesn't exist
            # However, this is bad UX, so we just return None
            self._value = None
            return None

    async def edit(self, note: Optional[str]) -> None:
        """|coro|

        Modifies the note. Can be at most 256 characters.

        Raises
        -------
        HTTPException
            Changing the note failed.
        """
        await self._state.http.set_note(self.user_id, note=note)
        self._value = note or ""

    async def delete(self) -> None:
        """|coro|

        A shortcut to :meth:`.edit` that deletes the note.

        Raises
        -------
        HTTPException
            Deleting the note failed.
        """
        await self.edit(None)

    def __repr__(self) -> str:
        base = f"<Note user={self.user!r}"
        note = self._value
        if note is not MISSING:
            note = note or ""
            return f"{base} note={note!r}>"
        return f"{base}>"

    def __str__(self) -> str:
        note = self._value
        if note is MISSING:
            raise ClientException("Note is not fetched")
        return note or ""

    def __bool__(self) -> bool:
        return bool(str(self))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Note) and self.user_id == other.user_id

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Note):
            return self._value != other._value or self.user_id != other.user_id
        return True

    def __hash__(self) -> int:
        return hash((self._value, self.user_id))

    def __len__(self) -> int:
        note = str(self)
        return len(note) if note else 0


class _UserTag:
    __slots__ = ()
    id: int


class BaseUser(_UserTag):
    __slots__ = (
        "name",
        "id",
        "discriminator",
        "_avatar",
        "_avatar_decoration",
        "_banner",
        "_accent_colour",
        "bot",
        "system",
        "_public_flags",
        "_cs_note",
        "_state",
    )

    if TYPE_CHECKING:
        name: str
        id: int
        discriminator: str
        bot: bool
        system: bool
        _state: ConnectionState
        _avatar: Optional[str]
        _avatar_decoration: Optional[str]
        _banner: Optional[str]
        _accent_colour: Optional[int]
        _public_flags: int

    def __init__(
        self, *, state: ConnectionState, data: Union[UserPayload, PartialUserPayload]
    ) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return (
            f"<BaseUser id={self.id} name={self.name!r} discriminator={self.discriminator!r}"
            f" bot={self.bot} system={self.system}>"
        )

    def __str__(self) -> str:
        return f"{self.name}#{self.discriminator}"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, _UserTag) and other.id == self.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.id >> 22

    def _update(self, data: Union[UserPayload, PartialUserPayload]) -> None:
        self.name = data["username"]
        self.id = int(data["id"])
        self.discriminator = data["discriminator"]
        self._avatar = data["avatar"]
        self._avatar_decoration = data.get("avatar_decoration")
        self._banner = data.get("banner", None)
        self._accent_colour = data.get("accent_color", None)
        self._public_flags = data.get("public_flags", 0)
        self.bot = data.get("bot", False)
        self.system = data.get("system", False)

    @classmethod
    def _copy(cls, user: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__

        self.name = user.name
        self.id = user.id
        self.discriminator = user.discriminator
        self._avatar = user._avatar
        self._avatar_decoration = user._avatar_decoration
        self._banner = user._banner
        self._accent_colour = user._accent_colour
        self._public_flags = user._public_flags
        self.bot = user.bot
        self.system = user.system
        self._state = user._state

        return self

    def _to_minimal_user_json(self) -> PartialUserPayload:
        user: PartialUserPayload = {
            "username": self.name,
            "id": self.id,
            "avatar": self._avatar,
            "avatar_decoration": self._avatar_decoration,
            "discriminator": self.discriminator,
            "bot": self.bot,
            "system": self.system,
            "public_flags": self._public_flags,
        }
        return user

    @property
    def voice(self) -> Optional[VoiceState]:
        """Optional[:class:`VoiceState`]: Returns the user's current voice state."""
        return self._state._voice_state_for(self.id)

    @property
    def public_flags(self) -> PublicUserFlags:
        """:class:`PublicUserFlags`: The publicly available flags the user has."""
        return PublicUserFlags._from_value(self._public_flags)

    @property
    def avatar(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the avatar the user has.

        If the user has not uploaded a global avatar, ``None`` is returned.
        If you want the avatar that a user has displayed, consider :attr:`display_avatar`.
        """
        if self._avatar is not None:
            return Asset._from_avatar(self._state, self.id, self._avatar)
        return None

    @property
    def default_avatar(self) -> Asset:
        """:class:`Asset`: Returns the default avatar for a given user. This is calculated by the user's discriminator."""
        return Asset._from_default_avatar(self._state, int(self.discriminator) % 5)

    @property
    def display_avatar(self) -> Asset:
        """:class:`Asset`: Returns the user's display avatar.

        For regular users this is just their default avatar or uploaded avatar.

        .. versionadded:: 2.0
        """
        return self.avatar or self.default_avatar

    @property
    def avatar_decoration(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the avatar decoration the user has.

        If the user does not have a avatar decoration, ``None`` is returned.

        .. versionadded:: 2.0
        """
        if self._avatar_decoration is not None:
            return Asset._from_avatar_decoration(
                self._state, self.id, self._avatar_decoration
            )
        return None

    @property
    def banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the user's banner asset, if available.

        .. versionadded:: 2.0

        .. note::
            This information is only available via :meth:`Client.fetch_user`.
        """
        if self._banner is None:
            return None
        return Asset._from_user_banner(self._state, self.id, self._banner)

    @property
    def display_banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the user's banner asset, if available.

        This is the same as :attr:`banner` and is here for compatibility.

        .. versionadded:: 2.0
        """
        return self.banner

    @property
    def accent_colour(self) -> Optional[Colour]:
        """Optional[:class:`Colour`]: Returns the user's accent colour, if applicable.

        A user's accent colour is only shown if they do not have a banner.
        This will only be available if the user explicitly sets a colour.

        There is an alias for this named :attr:`accent_color`.

        .. versionadded:: 2.0

        .. note::

            This information is only available via :meth:`Client.fetch_user`.
        """
        if self._accent_colour is None:
            return None
        return Colour(self._accent_colour)

    @property
    def accent_color(self) -> Optional[Colour]:
        """Optional[:class:`Colour`]: Returns the user's accent color, if applicable.

        A user's accent color is only shown if they do not have a banner.
        This will only be available if the user explicitly sets a color.

        There is an alias for this named :attr:`accent_colour`.

        .. versionadded:: 2.0

        .. note::

            This information is only available via :meth:`Client.fetch_user`.
        """
        return self.accent_colour

    @property
    def colour(self) -> Colour:
        """:class:`Colour`: A property that returns a colour denoting the rendered colour
        for the user. This always returns :meth:`Colour.default`.

        There is an alias for this named :attr:`color`.
        """
        return Colour.default()

    @property
    def color(self) -> Colour:
        """:class:`Colour`: A property that returns a color denoting the rendered color
        for the user. This always returns :meth:`Colour.default`.

        There is an alias for this named :attr:`colour`.
        """
        return self.colour

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string that allows you to mention the given user."""
        return f"<@{self.id}>"

    @property
    def created_at(self) -> datetime:
        """:class:`datetime.datetime`: Returns the user's creation time in UTC.

        This is when the user's Discord account was created.
        """
        return snowflake_time(self.id)

    @property
    def display_name(self) -> str:
        """:class:`str`: Returns the user's display name.

        For regular users this is just their username, but
        if they have a guild specific nickname then that
        is returned instead.
        """
        return self.name

    @cached_slot_property("_cs_note")
    def note(self) -> Note:
        """:class:`Note`: Returns an object representing the user's note.

        .. versionadded:: 2.0

        .. note::

            The underlying note is cached and updated from gateway events.
        """
        return Note(self._state, self.id, user=self)  # type: ignore

    def mentioned_in(self, message: Message) -> bool:
        """Checks if the user is mentioned in the specified message.

        Parameters
        -----------
        message: :class:`Message`
            The message to check if you're mentioned in.

        Returns
        -------
        :class:`bool`
            Indicates if the user is mentioned in the message.
        """
        if message.mention_everyone:
            return True

        return any(user.id == self.id for user in message.mentions)


class ClientUser(BaseUser):
    """Represents your Discord user.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: x != y

            Checks if two users are not equal.

        .. describe:: hash(x)

            Return the user's hash.

        .. describe:: str(x)

            Returns the user's name with discriminator.

    .. versionchanged:: 2.0
        :attr:`Locale` is now a :class:`Locale` instead of a Optional[:class:`str`].

    Attributes
    -----------
    name: :class:`str`
        The user's username.
    id: :class:`int`
        The user's unique ID.
    discriminator: :class:`str`
        The user's discriminator.
    bio: Optional[:class:`str`]
        The user's "about me" field. Could be ``None``.
    bot: :class:`bool`
        Specifies if the user is a bot account.
    system: :class:`bool`
        Specifies if the user is a system user (i.e. represents Discord officially).

        .. versionadded:: 1.3
    verified: :class:`bool`
        Specifies if the user's email is verified.
    email: Optional[:class:`str`]
        The email of the user.
    phone: Optional[:class:`int`]
        The phone number of the user.

        .. versionadded:: 1.9
    mfa_enabled: :class:`bool`
        Specifies if the user has MFA turned on and working.
    premium_type: Optional[:class:`PremiumType`]
        Specifies the type of premium a user has (i.e. Nitro, Nitro Classic, or Nitro Basic). Could be None if the user is not premium.
    note: :class:`Note`
        The user's note. Not pre-fetched.

        .. versionadded:: 1.9
    nsfw_allowed: Optional[:class:`bool`]
        Specifies if the user should be allowed to access NSFW content.
        If ``None``, then the user's date of birth is not known.

        .. versionadded:: 2.0
    desktop: :class:`bool`
        Specifies whether the user has used a desktop client.

        .. versionadded:: 2.0
    mobile: :class:`bool`
        Specifies whether the user has used a mobile client.

        .. versionadded:: 2.0
    """

    __slots__ = (
        "__weakref__",
        "_locale",
        "_flags",
        "verified",
        "mfa_enabled",
        "email",
        "phone",
        "premium_type",
        "note",
        "bio",
        "nsfw_allowed",
        "desktop",
        "mobile",
        "_purchased_flags",
        "_premium_usage_flags",
    )

    if TYPE_CHECKING:
        verified: bool
        email: Optional[str]
        phone: Optional[int]
        _locale: str
        _flags: int
        mfa_enabled: bool
        premium_type: Optional[PremiumType]
        bio: Optional[str]
        nsfw_allowed: Optional[bool]

    def __init__(self, *, state: ConnectionState, data: UserPayload) -> None:
        self._state = state
        self._full_update(data)
        self.note: Note = Note(state, self.id)

    def __repr__(self) -> str:
        return (
            f"<ClientUser id={self.id} name={self.name!r} discriminator={self.discriminator!r}"
            f" bot={self.bot} verified={self.verified} mfa_enabled={self.mfa_enabled} premium={self.premium}>"
        )

    def _full_update(self, data: UserPayload) -> None:
        self._update(data)
        self.verified = data.get("verified", False)
        self.email = data.get("email")
        self.phone = _get_as_snowflake(data, "phone")
        self._locale = data.get("locale", "en-US")
        self._flags = data.get("flags", 0)
        self._purchased_flags = data.get("purchased_flags", 0)
        self._premium_usage_flags = data.get("premium_usage_flags", 0)
        self.mfa_enabled = data.get("mfa_enabled", False)
        self.premium_type = (
            try_enum(PremiumType, data.get("premium_type"))
            if data.get("premium_type")
            else None
        )
        self.bio = data.get("bio") or None
        self.nsfw_allowed = data.get("nsfw_allowed")
        self.desktop: bool = data.get("desktop", False)
        self.mobile: bool = data.get("mobile", False)

    def _update_self(self, *args: Any) -> None:
        # ClientUser is kept up to date by USER_UPDATEs only
        return

    @property
    def locale(self) -> Locale:
        """:class:`Locale`: The IETF language tag used to identify the language the user is using."""
        return (
            self._state.settings.locale
            if self._state.settings
            else try_enum(Locale, self._locale)
        )

    @property
    def premium(self) -> bool:
        """Indicates if the user is a premium user (i.e. has Discord Nitro)."""
        return self.premium_type is not None

    @property
    def flags(self) -> PrivateUserFlags:
        """:class:`PrivateUserFlags`: Returns the user's flags (including private).

        .. versionadded:: 2.0
        """
        return PrivateUserFlags._from_value(self._flags)

    @property
    def premium_usage_flags(self) -> PremiumUsageFlags:
        """:class:`PremiumUsageFlags`: Returns the user's premium usage flags.

        .. versionadded:: 2.0
        """
        return PremiumUsageFlags._from_value(self._premium_usage_flags)

    @property
    def purchased_flags(self) -> PurchasedFlags:
        """:class:`PurchasedFlags`: Returns the user's purchased flags.

        .. versionadded:: 2.0
        """
        return PurchasedFlags._from_value(self._purchased_flags)

    async def edit(
        self,
        *,
        username: str = MISSING,
        avatar: Optional[bytes] = MISSING,
        avatar_decoration: Optional[bytes] = MISSING,
        password: str = MISSING,
        new_password: str = MISSING,
        email: str = MISSING,
        house: Optional[HypeSquadHouse] = MISSING,
        discriminator: Snowflake = MISSING,
        banner: Optional[bytes] = MISSING,
        accent_colour: Colour = MISSING,
        accent_color: Colour = MISSING,
        bio: Optional[str] = MISSING,
        date_of_birth: datetime = MISSING,
    ) -> ClientUser:
        """|coro|

        Edits the current profile of the client.

        .. note::

            To upload an avatar, a :term:`py:bytes-like object` must be passed in that
            represents the image being uploaded. If this is done through a file
            then the file must be opened via ``open('some_filename', 'rb')`` and
            the :term:`py:bytes-like object` is given through the use of ``fp.read()``.

        .. versionchanged:: 2.0
            The edit is no longer in-place, instead the newly edited client user is returned.

        .. versionchanged:: 2.0
            This function will now raise :exc:`ValueError` instead of
            ``InvalidArgument``.

        Parameters
        -----------
        password: :class:`str`
            The current password for the client's account.
            Required for everything except avatar, banner, accent_colour, date_of_birth, and bio.
        new_password: :class:`str`
            The new password you wish to change to.
        email: :class:`str`
            The new email you wish to change to.
        house: Optional[:class:`HypeSquadHouse`]
            The hypesquad house you wish to change to.
            Could be ``None`` to leave the current house.
        username: :class:`str`
            The new username you wish to change to.
        discriminator: :class:`int`
            The new discriminator you wish to change to.
            Can only be used if you have Nitro.
        avatar: Optional[:class:`bytes`]
            A :term:`py:bytes-like object` representing the image to upload.
            Could be ``None`` to denote no avatar.
        avatar_decoration: Optional[:class:`bytes`]
            A :term:`py:bytes-like object` representing the image to upload.
            Could be ``None`` to denote no avatar decoration.

            .. versionadded:: 2.0
        banner: :class:`bytes`
            A :term:`py:bytes-like object` representing the image to upload.
            Could be ``None`` to denote no banner.
        accent_colour: :class:`Colour`
            A :class:`Colour` object of the colour you want to set your profile to.

            .. versionadded:: 2.0
        bio: :class:`str`
            Your "about me" section.
            Could be ``None`` to represent no bio.

            .. versionadded:: 2.0
        date_of_birth: :class:`datetime.datetime`
            Your date of birth. Can only ever be set once.

            .. versionadded:: 2.0

        Raises
        ------
        HTTPException
            Editing your profile failed.
        ValueError
            Password was not passed when it was required.
            `house` field was not a :class:`HypeSquadHouse`.
            `date_of_birth` field was not a :class:`datetime.datetime`.
            `accent_colo(u)r` parameter was not a :class:`Colour`.

        Returns
        ---------
        :class:`ClientUser`
            The newly edited client user.
        """
        args: Dict[str, Any] = {}

        if any(
            x is not MISSING for x in (new_password, email, username, discriminator)
        ):
            if password is MISSING:
                raise ValueError("Password is required")
            args["password"] = password

        if avatar is not MISSING:
            if avatar is not None:
                args["avatar"] = _bytes_to_base64_data(avatar)
            else:
                args["avatar"] = None

        if avatar_decoration is not MISSING:
            if avatar_decoration is not None:
                args["avatar_decoration"] = _bytes_to_base64_data(avatar_decoration)
            else:
                args["avatar_decoration"] = None

        if banner is not MISSING:
            if banner is not None:
                args["banner"] = _bytes_to_base64_data(banner)
            else:
                args["banner"] = None

        if accent_color is not MISSING or accent_colour is not MISSING:
            colour = accent_colour if accent_colour is not MISSING else accent_color
            if colour is None:
                args["accent_color"] = colour
            elif not isinstance(colour, Colour):
                raise ValueError("`accent_colo(u)r` parameter was not a Colour")
            else:
                args["accent_color"] = accent_color.value

        if email is not MISSING:
            args["email"] = email

        if username is not MISSING:
            args["username"] = username

        if discriminator is not MISSING:
            args["discriminator"] = discriminator

        if new_password is not MISSING:
            args["new_password"] = new_password

        if bio is not MISSING:
            args["bio"] = bio or ""

        if date_of_birth is not MISSING:
            if not isinstance(date_of_birth, datetime):
                raise ValueError("`date_of_birth` parameter was not a datetime")
            args["date_of_birth"] = date_of_birth.strftime("%F")

        http = self._state.http

        if house is not MISSING:
            if house is None:
                await http.leave_hypesquad_house()
            elif not isinstance(house, HypeSquadHouse):
                raise ValueError("`house` parameter was not a HypeSquadHouse")
            else:
                await http.change_hypesquad_house(house.value)

        data = await http.edit_profile(args)
        try:
            http._token(data["token"])
        except KeyError:
            pass

        return ClientUser(state=self._state, data=data)


class User(BaseUser, discord.abc.Connectable, discord.abc.Messageable):
    """Represents a Discord user.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: x != y

            Checks if two users are not equal.

        .. describe:: hash(x)

            Return the user's hash.

        .. describe:: str(x)

            Returns the user's name with discriminator.

    Attributes
    -----------
    name: :class:`str`
        The user's username.
    id: :class:`int`
        The user's unique ID.
    discriminator: :class:`str`
        The user's discriminator.
    bot: :class:`bool`
        Specifies if the user is a bot account.
    system: :class:`bool`
        Specifies if the user is a system user (i.e. represents Discord officially).
    """

    __slots__ = ("__weakref__",)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r} discriminator={self.discriminator!r} bot={self.bot} system={self.system}>"

    def _get_voice_client_key(self) -> Tuple[int, str]:
        return self._state.self_id, "self_id"  # type: ignore # self_id is always set at this point

    def _get_voice_state_pair(self) -> Tuple[int, int]:
        return self._state.self_id, self.dm_channel.id  # type: ignore # self_id is always set at this point

    def _update_self(
        self, user: Union[PartialUserPayload, Tuple[()]]
    ) -> Optional[Tuple[User, User]]:
        if len(user) == 0 or len(user) <= 1:  # Done because of typing
            return

        original = (
            self.name,
            self._avatar,
            self.discriminator,
            self._public_flags,
            self._avatar_decoration,
        )
        # These keys seem to always be available
        modified = (
            user["username"],
            user.get("avatar"),
            user["discriminator"],
            user.get("public_flags", 0),
            user.get("avatar_decoration"),
        )
        if original != modified:
            to_return = User._copy(self)
            (
                self.name,
                self._avatar,
                self.discriminator,
                self._public_flags,
                self._avatar_decoration,
            ) = modified
            # Signal to dispatch user_update
            return to_return, self

    async def _get_channel(self) -> DMChannel:
        ch = await self.create_dm()
        return ch

    @property
    def dm_channel(self) -> Optional[DMChannel]:
        """Optional[:class:`DMChannel`]: Returns the channel associated with this user if it exists.

        If this returns ``None``, you can create a DM channel by calling the
        :meth:`create_dm` coroutine function.
        """
        return self._state._get_private_channel_by_user(self.id)

    @property
    def call(self) -> Optional[PrivateCall]:
        """Optional[:class:`PrivateCall`]: Returns the call associated with this user if it exists."""
        return getattr(self.dm_channel, "call", None)

    @property
    def relationship(self) -> Optional[Relationship]:
        """Optional[:class:`Relationship`]: Returns the :class:`Relationship` with this user if applicable, ``None`` otherwise."""
        return self._state._relationships.get(self.id)

    @copy_doc(discord.abc.Connectable.connect)
    async def connect(
        self,
        *,
        timeout: float = 60.0,
        reconnect: bool = True,
        cls: Callable[[Client, VocalChannel], ConnectReturn] = VoiceClient,
        ring: bool = True,
    ) -> ConnectReturn:
        channel = await self._get_channel()
        call = channel.call
        if call is None and ring:
            await channel._initial_ring()
        return await super().connect(
            timeout=timeout, reconnect=reconnect, cls=cls, _channel=channel
        )

    async def create_dm(self) -> DMChannel:
        """|coro|

        Creates a :class:`DMChannel` with this user.

        This should be rarely called, as this is done transparently for most
        people.

        Returns
        -------
        :class:`.DMChannel`
            The channel that was created.
        """
        found = self.dm_channel
        if found is not None:
            return found

        state = self._state
        data: DMChannelPayload = await state.http.start_private_message(self.id)
        return state.add_dm_channel(data)

    def is_friend(self) -> bool:
        """:class:`bool`: Checks if the user is your friend."""
        r = self.relationship
        if r is None:
            return False
        return r.type is RelationshipType.friend

    def is_blocked(self) -> bool:
        """:class:`bool`: Checks if the user is blocked."""
        r = self.relationship
        if r is None:
            return False
        return r.type is RelationshipType.blocked

    async def block(self) -> None:  # TODO: maybe return relationship
        """|coro|

        Blocks the user.

        Raises
        -------
        Forbidden
            Not allowed to block this user.
        HTTPException
            Blocking the user failed.
        """
        await self._state.http.add_relationship(
            self.id,
            type=RelationshipType.blocked.value,
            action=RelationshipAction.block,
        )

    async def unblock(self) -> None:
        """|coro|

        Unblocks the user.

        Raises
        -------
        Forbidden
            Not allowed to unblock this user.
        HTTPException
            Unblocking the user failed.
        """
        await self._state.http.remove_relationship(
            self.id, action=RelationshipAction.unblock
        )

    async def remove_friend(self) -> None:
        """|coro|

        Removes the user as a friend.

        Raises
        -------
        Forbidden
            Not allowed to remove this user as a friend.
        HTTPException
            Removing the user as a friend failed.
        """
        await self._state.http.remove_relationship(
            self.id, action=RelationshipAction.unfriend
        )

    async def send_friend_request(self) -> None:
        """|coro|

        Sends the user a friend request.

        Raises
        -------
        Forbidden
            Not allowed to send a friend request to the user.
        HTTPException
            Sending the friend request failed.
        """
        await self._state.http.send_friend_request(self.name, self.discriminator)

    async def profile(
        self,
        *,
        with_mutual_guilds: bool = True,
        with_mutual_friends_count: bool = False,
        with_mutual_friends: bool = True,
    ) -> UserProfile:
        """|coro|

        A shorthand method to retrieve a :class:`UserProfile` for the user.

        Parameters
        ------------
        with_mutual_guilds: :class:`bool`
            Whether to fetch mutual guilds.
            This fills in :attr:`UserProfile.mutual_guilds`.

            .. versionadded:: 2.0
        with_mutual_friends_count: :class:`bool`
            Whether to fetch the number of mutual friends.
            This fills in :attr:`UserProfile.mutual_friends_count`.

            .. versionadded:: 2.0
        with_mutual_friends: :class:`bool`
            Whether to fetch mutual friends.
            This fills in :attr:`UserProfile.mutual_friends` and :attr:`UserProfile.mutual_friends_count`,
            but requires an extra API call.

            .. versionadded:: 2.0

        Raises
        -------
        Forbidden
            Not allowed to fetch this profile.
        HTTPException
            Fetching the profile failed.

        Returns
        --------
        :class:`UserProfile`
            The profile of the user.
        """
        return await self._state.client.fetch_user_profile(
            self.id,
            with_mutual_guilds=with_mutual_guilds,
            with_mutual_friends_count=with_mutual_friends_count,
            with_mutual_friends=with_mutual_friends,
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional

from .application import (
    ApplicationActivityStatistics,
    ApplicationBranch,
    PartialApplication,
)
from .entitlements import Entitlement
from .enums import SKUType, try_enum
from .flags import LibraryApplicationFlags
from .mixins import Hashable
from .utils import MISSING, _get_as_snowflake, find, parse_date, parse_time

if TYPE_CHECKING:
    from datetime import date, datetime

    from .asset import Asset
    from .state import ConnectionState
    from .types.application import Branch as BranchPayload
    from .types.library import LibraryApplication as LibraryApplicationPayload
    from .types.store import PartialSKU as PartialSKUPayload

__all__ = (
    "LibrarySKU",
    "LibraryApplication",
)


class LibrarySKU(Hashable):
    """Represents a partial store SKU for a library entry.

    .. container:: operations

        .. describe:: x == y

            Checks if two library SKUs are equal.

        .. describe:: x != y

            Checks if two library SKUs are not equal.

        .. describe:: hash(x)

            Returns the library SKU's hash.

    .. versionadded:: 2.0

    Attributes
    -----------
    id: :class:`int`
        The SKU's ID.
    type: :class:`SKUType`
        The type of the SKU.
    preorder_release_date: Optional[:class:`datetime.date`]
        The approximate date that the SKU will released for pre-order, if any.
    preorder_released_at: Optional[:class:`datetime.datetime`]
        The date that the SKU was released for pre-order, if any.
    premium: :class:`bool`
        Whether this SKU is provided for free to premium users.
    """

    __slots__ = (
        "id",
        "type",
        "preorder_release_date",
        "preorder_released_at",
        "premium",
    )

    def __init__(self, data: PartialSKUPayload):
        self.id: int = int(data["id"])
        self.type: SKUType = try_enum(SKUType, data["type"])
        self.preorder_release_date: Optional[date] = parse_date(
            data.get("preorder_approximate_release_date")
        )
        self.preorder_released_at: Optional[datetime] = parse_time(
            data.get("preorder_release_at")
        )
        self.premium: bool = data.get("premium", False)

    def __repr__(self) -> str:
        return f"<LibrarySKU id={self.id} type={self.type!r} preorder_release_date={self.preorder_release_date!r} preorder_released_at={self.preorder_released_at!r} premium={self.premium!r}>"


class LibraryApplication:
    """Represents a library entry.

    .. container:: operations

        .. describe:: x == y

            Checks if two library entries are equal.

        .. describe:: x != y

            Checks if two library entries are not equal.

        .. describe:: hash(x)

            Returns the library entry's hash.

        .. describe:: str(x)

            Returns the library entry's name.

    .. versionadded:: 2.0

    Attributes
    -----------
    created_at: :class:`datetime.datetime`
        When this library entry was created.
    application: :class:`PartialApplication`
        The application that this library entry is for.
    sku_id: :class:`int`
        The ID of the SKU that this library entry is for.
    sku: :class:`LibrarySKU`
        The SKU that this library entry is for.
    entitlements: List[:class:`Entitlement`]
        The entitlements that this library entry has.
    branch_id: :class:`int`
        The ID of the branch that this library entry installs.
    branch: :class:`ApplicationBranch`
        The branch that this library entry installs.
    """

    __slots__ = (
        "created_at",
        "application",
        "sku_id",
        "sku",
        "entitlements",
        "branch_id",
        "branch",
        "_flags",
        "_state",
    )

    def __init__(self, *, state: ConnectionState, data: LibraryApplicationPayload):
        self._state = state
        self._update(data)

    def _update(self, data: LibraryApplicationPayload):
        state = self._state

        self.created_at: datetime = parse_time(data["created_at"])
        self.application: PartialApplication = PartialApplication(
            state=state, data=data["application"]
        )
        self.sku_id: int = int(data["sku_id"])
        self.sku: LibrarySKU = LibrarySKU(data=data["sku"])
        self.entitlements: List[Entitlement] = [
            Entitlement(state=state, data=e) for e in data.get("entitlements", [])
        ]
        self._flags = data.get("flags", 0)

        self.branch_id: int = int(data["branch_id"])
        branch: Optional[BranchPayload] = data.get("branch")
        if not branch:
            branch = {"id": self.branch_id, "name": "master"}
        self.branch: ApplicationBranch = ApplicationBranch(
            state=state, data=branch, application_id=self.application.id
        )

    def __repr__(self) -> str:
        return f"<LibraryApplication created_at={self.created_at!r} application={self.application!r} sku={self.sku!r} branch={self.branch!r}>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, LibraryApplication):
            return (
                self.application.id == other.application.id
                and self.branch_id == other.branch_id
            )
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, LibraryApplication):
            return (
                self.application.id != other.application.id
                or self.branch_id != other.branch_id
            )
        return True

    def __hash__(self) -> int:
        return hash((self.application.id, self.branch_id))

    def __str__(self) -> str:
        return self.application.name

    @property
    def name(self) -> str:
        """:class:`str`: The library entry's name."""
        return self.application.name

    @property
    def icon(self) -> Optional[Asset]:
        """:class:`Asset`: The library entry's icon asset, if any."""
        return self.application.icon

    @property
    def flags(self) -> LibraryApplicationFlags:
        """:class:`LibraryApplicationFlags`: The library entry's flags."""
        return LibraryApplicationFlags._from_value(self._flags)

    async def activity_statistics(self) -> ApplicationActivityStatistics:
        """|coro|

        Gets the activity statistics for this library entry.

        Raises
        -------
        HTTPException
            Getting the activity statistics failed.

        Returns
        --------
        :class:`ApplicationActivityStatistics`
            The activity statistics for this library entry.
        """
        state = self._state
        data = await state.http.get_activity_statistics()
        app = find(
            lambda a: _get_as_snowflake(a, "application_id") == self.application.id,
            data,
        )
        return ApplicationActivityStatistics(
            data=app
            or {
                "application_id": self.application.id,
                "total_duration": 0,
                "last_played_at": "1970-01-01T00:00:00+00:00",
            },
            state=state,
        )

    async def mark_installed(self) -> None:
        """|coro|

        Marks the library entry as installed.

        Raises
        -------
        HTTPException
            Marking the library entry as installed failed.
        """
        await self._state.http.mark_library_entry_installed(
            self.application.id, self.branch_id
        )

    async def edit(self, *, flags: LibraryApplicationFlags = MISSING) -> None:
        """|coro|

        Edits the library entry.

        All parameters are optional.

        Parameters
        -----------
        flags: :class:`LibraryApplicationFlags`
            The new flags to set for the library entry.

        Raises
        -------
        HTTPException
            Editing the library entry failed.
        """
        payload = {}
        if flags is not MISSING:
            payload["flags"] = flags.value

        data = await self._state.http.edit_library_entry(
            self.application.id, self.branch_id, payload
        )
        self._update(data)

    async def delete(self) -> None:
        """|coro|

        Deletes the library entry.

        Raises
        -------
        HTTPException
            Deleting the library entry failed.
        """
        await self._state.http.delete_library_entry(self.application.id, self.branch_id)



from discord_self._vendor.discord.types.member import MemberWithUser
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.voice import (
    GuildVoiceState,
    SupportedModes,
    VoiceIdentify,
    VoiceReady,
    VoiceRegion,
    VoiceServerUpdate,
    VoiceState,
)

__all__ = [
    "GuildVoiceState",
    "MemberWithUser",
    "Snowflake",
    "SupportedModes",
    "VoiceIdentify",
    "VoiceReady",
    "VoiceRegion",
    "VoiceServerUpdate",
    "VoiceState",
]



from discord_self._vendor.discord.errors import (
    AuthFailure,
    CaptchaRequired,
    ClientException,
    ConnectionClosed,
    DiscordException,
    DiscordServerError,
    Forbidden,
    GatewayNotFound,
    HTTPException,
    InvalidData,
    LoginFailure,
    NotFound,
    RateLimited,
)

__all__ = [
    "AuthFailure",
    "CaptchaRequired",
    "ClientException",
    "ConnectionClosed",
    "DiscordException",
    "DiscordServerError",
    "Forbidden",
    "GatewayNotFound",
    "HTTPException",
    "InvalidData",
    "LoginFailure",
    "NotFound",
    "RateLimited",
]



from discord_self._vendor.discord.profile import (
    ApplicationProfile,
    MemberProfile,
    MutualGuild,
    UserProfile,
)

__all__ = ["ApplicationProfile", "MemberProfile", "MutualGuild", "UserProfile"]



from discord_self._vendor.discord.sticker import (
    GuildSticker,
    StandardSticker,
    Sticker,
    StickerItem,
    StickerPack,
)

__all__ = ["GuildSticker", "StandardSticker", "Sticker", "StickerItem", "StickerPack"]
